from __future__ import annotations

from collections.abc import MutableMapping
from copy import copy
from typing import TYPE_CHECKING
from typing import cast
from warnings import warn

import pytest

from glotaran.builtin.io.sdt.sdt_file_reader import SdtDataIo
from glotaran.builtin.io.yml.yml import YmlProjectIo
from glotaran.builtin.megacomplexes.decay import DecayMegacomplex
from glotaran.io.interface import DataIoInterface
from glotaran.io.interface import ProjectIoInterface
from glotaran.model.megacomplex import Megacomplex
from glotaran.plugin_system.base_registry import PluginOverwriteWarning
from glotaran.plugin_system.base_registry import add_instantiated_plugin_to_registry
from glotaran.plugin_system.base_registry import add_plugin_to_registry
from glotaran.plugin_system.base_registry import full_plugin_name
from glotaran.plugin_system.base_registry import get_method_from_plugin
from glotaran.plugin_system.base_registry import get_plugin_from_registry
from glotaran.plugin_system.base_registry import is_registered_plugin
from glotaran.plugin_system.base_registry import methods_differ_from_baseclass
from glotaran.plugin_system.base_registry import methods_differ_from_baseclass_table
from glotaran.plugin_system.base_registry import registered_plugins
from glotaran.plugin_system.base_registry import set_plugin
from glotaran.plugin_system.base_registry import show_method_help
from glotaran.plugin_system.base_registry import supported_file_extensions

if TYPE_CHECKING:
    from _pytest.capture import CaptureFixture

    from glotaran.plugin_system.base_registry import _PluginInstantiableType
    from glotaran.plugin_system.base_registry import _PluginType


class MockPlugin:
    format_name = "mock"

    def some_method(self):
        """This docstring is just for help testing of 'some_method'."""
        return f"got the method of {self.format_name}"

    def some_other_method(self):
        pass


class MockPluginSubclassPartial(MockPlugin):
    def some_method(self):
        return "different implementation"


class MockPluginSubclassFull(MockPlugin):
    def some_method(self):
        return "different implementation"

    def some_other_method(self):
        return "different implementation"


class MockPluginSubclassStr(MockPlugin):
    def some_method(self):
        return "different implementation"


def get_mock_plugin_function(plugin_registry_key: str):
    if plugin_registry_key == "base":
        return MockPlugin
    elif plugin_registry_key == "sub_class_partial":
        return MockPluginSubclassPartial
    elif plugin_registry_key == "sub_class_full":
        return MockPluginSubclassFull
    elif plugin_registry_key == "sub_class_str":
        return MockPluginSubclassStr
    raise ValueError(f"No mock plugin with name {plugin_registry_key!r}")


mock_registry_data_io = cast(
    MutableMapping[str, DataIoInterface],
    {
        "sdt": SdtDataIo("sdt"),
        "yml": YmlProjectIo("yml"),
        "glotaran.builtin.io.sdt.sdt_file_reader.SdtDataIo": SdtDataIo("sdt"),
        "decay": DecayMegacomplex,
        "mock_plugin": MockPlugin,
        "imported_plugin": MockPlugin,
    },
)

mock_registry_project_io = cast(
    MutableMapping[str, ProjectIoInterface], copy(mock_registry_data_io)
)
mock_registry_model = cast(MutableMapping[str, type[Megacomplex]], copy(mock_registry_data_io))


@pytest.mark.parametrize(
    "plugin, expected",
    (
        (SdtDataIo, "glotaran.builtin.io.sdt.sdt_file_reader.SdtDataIo"),
        (SdtDataIo("sdt"), "glotaran.builtin.io.sdt.sdt_file_reader.SdtDataIo"),
    ),
)
def test_full_plugin_name(plugin: object | type[object], expected: str):
    """Full name of class and instance is retrieved correctly."""
    assert full_plugin_name(plugin) == expected


def test_PluginOverwriteWarning():
    """Ignore the values this is just for testing, even if it doesn't make sense."""

    with pytest.warns(PluginOverwriteWarning) as record:
        warn(
            PluginOverwriteWarning(
                old_key="yml",
                old_plugin=YmlProjectIo,
                new_plugin=SdtDataIo,
                plugin_set_func_name="set_plugin",
            )
        )

        assert len(record) == 1
        expected = (
            "The plugin 'glotaran.builtin.io.sdt.sdt_file_reader.SdtDataIo' tried to "
            "overwrite the plugin 'glotaran.builtin.io.yml.yml.YmlProjectIo', "
            "with the access_name 'yml'. "
            "Use set_plugin('yml', 'glotaran.builtin.io.sdt.sdt_file_reader.SdtDataIo') "
            "to use 'glotaran.builtin.io.sdt.sdt_file_reader.SdtDataIo' instead."
        )

        assert record[0].message.args[0] == expected


@pytest.mark.parametrize(
    "plugin_register_key, plugin, plugin_registry",
    (
        ("sdt_new", SdtDataIo("sdt"), copy(mock_registry_data_io)),
        ("yml_new", YmlProjectIo("yml"), copy(mock_registry_project_io)),
        ("decay_new", DecayMegacomplex, copy(mock_registry_model)),
    ),
)
def test_add_plugin_to_register(
    plugin_register_key: str,
    plugin: _PluginType,
    plugin_registry: MutableMapping[str, _PluginType],
):
    """Add plugin with one key"""
    add_plugin_to_registry(plugin_register_key, plugin, plugin_registry, "set_plugin")
    assert plugin_register_key in plugin_registry
    assert plugin_registry[plugin_register_key] == plugin


def test_add_plugin_to_register_naming_error():
    """Raise error if pluginkey contains '.'"""
    with pytest.raises(
        ValueError,
        match=(
            r"The character '\.' isn't allowed in the name of a plugin, "
            r"you provided the name 'bad\.name'."
        ),
    ):
        add_plugin_to_registry("bad.name", MockPlugin, mock_registry_model, "set_plugin")


def test_add_plugin_to_register_existing_plugin():
    """Warn if different plugin overwrites existing."""

    plugin_registry = copy(mock_registry_data_io)
    with pytest.warns(PluginOverwriteWarning, match="SdtDataIo.+sdt.+YmlProjectIo") as record:
        add_plugin_to_registry("sdt", YmlProjectIo("sdt"), plugin_registry, "set_plugin")
        assert len(record) == 1
    assert plugin_registry["glotaran.builtin.io.yml.yml.YmlProjectIo"].format == "sdt"


@pytest.mark.xfail(strict=True, reason="Should not warn")
def test_add_plugin_to_register_existing_plugin_self():
    """Don't warn if plugin overwrites itself."""

    plugin_registry = copy(mock_registry_data_io)
    with pytest.warns(PluginOverwriteWarning, match="SdtDataIo.+sdt.+SdtDataIo") as record:
        add_plugin_to_registry("sdt", SdtDataIo("sdt"), plugin_registry, "set_plugin")
        assert len(record) == 0


@pytest.mark.parametrize(
    "plugin_register_key, plugin, plugin_registry",
    (
        ("sdt_new", SdtDataIo, copy(mock_registry_data_io)),
        ("yml_new", YmlProjectIo, copy(mock_registry_project_io)),
    ),
)
def test_add_instantiated_plugin_to_register(
    plugin_register_key: str,
    plugin: type[_PluginInstantiableType],
    plugin_registry: MutableMapping[str, _PluginInstantiableType],
):
    """Add instantiated plugin"""
    add_instantiated_plugin_to_registry(plugin_register_key, plugin, plugin_registry, "set_plugin")
    assert plugin_register_key in plugin_registry
    assert plugin_registry[plugin_register_key].format == plugin_register_key


@pytest.mark.parametrize(
    "plugin_register_keys, plugin, plugin_registry",
    (
        (["sdt_new", "sdt_new2"], SdtDataIo, copy(mock_registry_data_io)),
        (["yml_new", "yaml_new"], YmlProjectIo, copy(mock_registry_project_io)),
    ),
)
def test_add_instantiated_plugin_to_register_multiple_keys(
    plugin_register_keys: list[str],
    plugin: type[_PluginInstantiableType],
    plugin_registry: MutableMapping[str, _PluginInstantiableType],
):
    """Add instantiated plugin with multiple keys"""
    add_instantiated_plugin_to_registry(
        plugin_register_keys, plugin, plugin_registry, "set_plugin"
    )
    for plugin_register_key in plugin_register_keys:
        assert plugin_register_key in plugin_registry
        assert plugin_registry[plugin_register_key].format == plugin_register_key


def test_set_plugin():
    """Overwrite existing plugin"""
    plugin_registry = copy(mock_registry_data_io)
    set_plugin("yml", "glotaran.builtin.io.sdt.sdt_file_reader.SdtDataIo", plugin_registry)
    assert isinstance(plugin_registry["yml"], SdtDataIo)


def test_set_plugin_dot_in_plugin_key():
    """Raise error if plugin_register_key contains '.'"""
    plugin_registry = copy(mock_registry_data_io)
    with pytest.raises(
        ValueError,
        match=r"The value of 'format_name' isn't allowed to contain the character '\.' \.",
    ):
        set_plugin("foo.bar", "glotaran.builtin.io.sdt.sdt_file_reader.SdtDataIo", plugin_registry)


def test_set_plugin_plugin_not_registered():
    """Raise error if the is not plugin"""
    plugin_registry = copy(mock_registry_data_io)
    with pytest.raises(
        ValueError,
        match=(
            r"There isn't a plugin registered under the full name 'mymodul\.NotRegistered'\.\n"
            r"Maybe you need to install a plugin\? Known plugins are:\n "
            r"\['glotaran\.builtin\.io\.sdt\.sdt_file_reader\.SdtDataIo'\]"
        ),
    ):
        set_plugin("yml", "mymodul.NotRegistered", plugin_registry)


def test_set_plugin_no_dot_full_name():
    """Raise error if full_plugin_name doesn't contains '.'"""
    plugin_registry = copy(mock_registry_data_io)
    with pytest.raises(
        ValueError,
        match=(
            r"There isn't a plugin registered under the full name 'sdt'\.\n"
            r"Maybe you need to install a plugin\? Known plugins are:\n "
            r"\['glotaran\.builtin\.io\.sdt\.sdt_file_reader\.SdtDataIo'\]"
        ),
    ):
        set_plugin("yml", "sdt", plugin_registry)


def test_registered_plugins():
    """List of registered names"""
    result = [
        "sdt",
        "yml",
        "decay",
        "mock_plugin",
        "imported_plugin",
    ]
    assert registered_plugins(mock_registry_data_io) == sorted(result)


def test_is_registered_plugin():
    """Registered name in registry"""
    assert is_registered_plugin("sdt", mock_registry_data_io)
    assert not is_registered_plugin("not-registered", mock_registry_data_io)


def test_get_plugin_from_register():
    """Retrieve plugin from registry"""
    plugin = get_plugin_from_registry("sdt", mock_registry_data_io, "something went wrong")
    assert plugin.format == "sdt"


def test_get_plugin_from_register_not_found():
    """Error when Plugin wasn't found"""
    with pytest.raises(ValueError, match="something went wrong"):
        get_plugin_from_registry("not-registered", mock_registry_data_io, "something went wrong")


def test_get_method_from_plugin():
    """Method works like a function."""
    plugin = MockPlugin()

    method = get_method_from_plugin(plugin, "some_method")

    assert method() == "got the method of mock"


def test_get_method_from_plugin_just_attribute():
    """Error if attribute does exists on plugin but is not a method."""
    plugin = MockPlugin()

    with pytest.raises(
        ValueError, match=r"The plugin '.+?\.MockPlugin' has no method 'format_name'"
    ):
        get_method_from_plugin(plugin, "format_name")


def test_get_method_from_plugin_missing_attribute():
    """Error if attribute does not exists on plugin."""
    plugin = MockPlugin()

    with pytest.raises(
        ValueError, match=r"The plugin '.+?\.MockPlugin' has no method 'not_even_an_attribute'"
    ):
        get_method_from_plugin(plugin, "not_even_an_attribute")


@pytest.mark.parametrize("plugin", (MockPlugin, MockPlugin()))
def test_show_method_help(capsys: CaptureFixture, plugin: MockPlugin | type[MockPlugin]):
    """Same help as when called directly.

    Note: help differs when called on class.method vs. instance.method
    """
    help(plugin.some_method)
    original_help, _ = capsys.readouterr()

    show_method_help(plugin, "some_method")
    result, _ = capsys.readouterr()

    assert "This docstring is just for help testing of 'some_method'." in result
    assert result == original_help


@pytest.mark.parametrize(
    "method_names,plugin,expected",
    (
        ("some_method", MockPluginSubclassPartial, [True]),
        ("some_method", MockPluginSubclassPartial(), [True]),
        (["some_method", "some_other_method"], MockPluginSubclassPartial, [True, False]),
        (["some_method", "some_other_method"], MockPluginSubclassPartial(), [True, False]),
        (["some_method", "some_other_method"], MockPluginSubclassFull, [True, True]),
        (["some_method", "some_other_method"], MockPluginSubclassFull(), [True, True]),
    ),
)
def test_methods_differ_from_baseclass(
    method_names: str | list[str], plugin: object | type[object], expected: list[bool]
):
    """Inherited methods are the same as base class and overwritten ones differ"""
    result = methods_differ_from_baseclass(method_names, plugin, MockPlugin)

    assert list(result) == expected


@pytest.mark.parametrize(
    "method_names,plugin_registry_keys,expected",
    (
        ("some_method", "base", [["`base`", False]]),
        ("some_method", "sub_class_partial", [["`sub_class_partial`", True]]),
        ("some_method", "sub_class_full", [["`sub_class_full`", True]]),
        (
            ["some_method", "some_other_method"],
            ["base", "sub_class_partial", "sub_class_full", "sub_class_str"],
            [
                ["`base`", False, False],
                ["`sub_class_partial`", True, False],
                ["`sub_class_full`", True, True],
                ["`sub_class_str`", True, False],
            ],
        ),
    ),
)
def test_methods_differ_from_baseclass_table(
    method_names: str | list[str],
    plugin_registry_keys: str | list[str],
    expected: list[list[str | bool]],
):
    """Inherited methods are the same as base class and overwritten ones differ"""

    result = methods_differ_from_baseclass_table(
        method_names, plugin_registry_keys, get_mock_plugin_function, MockPlugin
    )

    assert list(result) == expected


def test_methods_differ_from_baseclass_table_plugin_names():
    """Show plugin name"""

    result = methods_differ_from_baseclass_table(
        "some_method", "base", get_mock_plugin_function, MockPlugin, plugin_names=True
    )

    assert list(result) == [["`base`", False, "`test_base_registry.MockPlugin`"]]


@pytest.mark.parametrize(
    "method_names, expected",
    (
        ("some_method", [".sub_class_partial", ".sub_class_full"]),
        ("some_other_method", [".sub_class_full"]),
        (["some_method", "some_other_method"], [".sub_class_full"]),
    ),
)
def test_supported_file_extensions(
    method_names: str | list[str],
    expected: list[str],
):
    """Only extensions where the plugin supports all methods in ``method_names`` are returned."""
    result = supported_file_extensions(
        method_names,
        ["base", "sub_class_partial", "sub_class_full", "sub_class_str"],
        get_mock_plugin_function,
        MockPlugin,
    )

    assert list(result) == expected



"""Baseclasses to create Data/Project IO plugins from.

The main purpose of those classes are to guarantee a consistent API via
typechecker like ``mypy`` and demonstarate with methods are accessed by
highlevel convenience functions for a given type of plugin.

To add additional options to a method, those options need to be
keyword only arguments.
See: https://www.python.org/dev/peps/pep-3102/

"""

from __future__ import annotations

from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from collections.abc import Callable
    from typing import Literal
    from typing import Union

    import xarray as xr

    from glotaran.model import Model
    from glotaran.parameter import Parameters
    from glotaran.project import Result
    from glotaran.project import Scheme

    DataLoader = Callable[[str], Union[xr.Dataset, xr.DataArray]]
    DataSaver = Callable[[str, Union[xr.Dataset, xr.DataArray]], None]


@dataclass
class SavingOptions:
    """A collection of options for result saving."""

    data_filter: list[str] | None = None
    data_format: Literal["nc"] = "nc"
    parameter_format: Literal["csv"] = "csv"
    report: bool = True


SAVING_OPTIONS_DEFAULT = SavingOptions()
SAVING_OPTIONS_MINIMAL = SavingOptions(data_filter=["fitted_data", "residual"], report=False)


class DataIoInterface:
    """Baseclass for Data IO plugins."""

    def __init__(self, format_name: str) -> None:
        """Initialize a Data IO plugin with the name of the format.

        Parameters
        ----------
        format_name: str
            Name of the supported format an instance uses.
        """
        self.format = format_name

    def load_dataset(self, file_name: str) -> xr.Dataset | xr.DataArray:
        """Read data from a file to :xarraydoc:`Dataset` or :xarraydoc:`DataArray`.

        **NOT IMPLEMENTED**

        Parameters
        ----------
        file_name: str
            File containing the data.

        Returns
        -------
        xr.Dataset|xr.DataArray
            Data loaded from the file.


        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"""Cannot read data with format: {self.format!r}""")

    def save_dataset(
        self,
        dataset: xr.Dataset | xr.DataArray,
        file_name: str,
    ):
        """Save data from :xarraydoc:`Dataset` to a file.

        **NOT IMPLEMENTED**

        Parameters
        ----------
        dataset: xr.Dataset
            Dataset to be saved to file.
        file_name: str
            File to write the data to.


        .. # noqa: DAR101
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"""Cannot save data with format: {self.format!r}""")


class ProjectIoInterface:
    """Baseclass for Project IO plugins."""

    def __init__(self, format_name: str) -> None:
        """Initialize a Project IO plugin with the name of the format.

        Parameters
        ----------
        format_name: str
            Name of the supported format an instance uses.
        """
        self.format = format_name

    def load_model(self, file_name: str) -> Model:
        """Create a Model instance from the specs defined in a file.

        **NOT IMPLEMENTED**

        Parameters
        ----------
        file_name: str
            File containing the model specs.

        Returns
        -------
        Model
            Model instance created from the file.


        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot read models with format {self.format!r}")

    def save_model(self, model: Model, file_name: str):
        """Save a Model instance to a spec file.

        **NOT IMPLEMENTED**

        Parameters
        ----------
        model: Model
            Model instance to save to specs file.
        file_name: str
            File to write the model specs to.


        .. # noqa: DAR101
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot save models with format {self.format!r}")

    def load_parameters(self, file_name: str) -> Parameters:
        """Create a Parameters instance from the specs defined in a file.

        **NOT IMPLEMENTED**

        Parameters
        ----------
        file_name: str
            File containing the parameter specs.

        Returns
        -------
        Parameters
            Parameters instance created from the file.


        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot read parameters with format {self.format!r}")

    def save_parameters(self, parameters: Parameters, file_name: str):
        """Save a Parameters instance to a spec file.

        **NOT IMPLEMENTED**

        Parameters
        ----------
        parameters: Parameters
            Parameters instance to save to specs file.
        file_name: str
            File to write the parameter specs to.


        .. # noqa: DAR101
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot save parameters with format {self.format!r}")

    def load_scheme(self, file_name: str) -> Scheme:
        """Create a Scheme instance from the specs defined in a file.

        **NOT IMPLEMENTED**

        Parameters
        ----------
        file_name: str
            File containing the parameter specs.

        Returns
        -------
        Scheme
            Scheme instance created from the file.

        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot read scheme with format {self.format!r}")

    def save_scheme(self, scheme: Scheme, file_name: str):
        """Save a Scheme instance to a spec file.

        **NOT IMPLEMENTED**

        Parameters
        ----------
        scheme: Scheme
            Scheme instance to save to specs file.
        file_name: str
            File to write the scheme specs to.


        .. # noqa: DAR101
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot save scheme with format {self.format!r}")

    def load_result(self, result_path: str) -> Result:
        """Create a Result instance from the specs defined in a file.

        **NOT IMPLEMENTED**

        Parameters
        ----------
        result_path: str
            Path containing the result data.

        Returns
        -------
        Result
            Result instance created from the file.


        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot read result with format {self.format!r}")

    def save_result(
        self,
        result: Result,
        result_path: str,
        *,
        saving_options: SavingOptions = SAVING_OPTIONS_DEFAULT,
    ) -> list[str]:
        """Save a Result instance to a spec file.

        **NOT IMPLEMENTED**

        Parameters
        ----------
        result: Result
            Result instance to save to specs file.
        result_path: str
            Path to write the result data to.
        saving_options: SavingOptions
            Options for the saved result.


        .. # noqa: DAR101
        .. # noqa: DAR401
        """
        raise NotImplementedError(f"Cannot save result with format {self.format!r}")



"""This package contains the kinetic megacomplex item."""
from __future__ import annotations

from typing import TYPE_CHECKING

import numba as nb
import numpy as np
import xarray as xr

from glotaran.builtin.megacomplexes.decay.decay_parallel_megacomplex import DecayDatasetModel
from glotaran.builtin.megacomplexes.decay.irf import IrfMultiGaussian
from glotaran.builtin.megacomplexes.decay.util import index_dependent
from glotaran.builtin.megacomplexes.decay.util import retrieve_irf
from glotaran.model import DatasetModel
from glotaran.model import Megacomplex
from glotaran.model import ModelError
from glotaran.model import ParameterType
from glotaran.model import megacomplex

if TYPE_CHECKING:
    from glotaran.typing.types import ArrayLike


@megacomplex(dataset_model_type=DecayDatasetModel, unique=True)
class CoherentArtifactMegacomplex(Megacomplex):
    dimension: str = "time"
    type: str = "coherent-artifact"
    order: int
    width: ParameterType | None = None

    def calculate_matrix(
        self,
        dataset_model: DatasetModel,
        global_axis: ArrayLike,
        model_axis: ArrayLike,
        **kwargs,
    ):
        if not 1 <= self.order <= 3:
            raise ModelError("Coherent artifact order must be between in [1,3]")

        irf = dataset_model.irf
        if irf is None:
            raise ModelError(f'No irf in dataset "{dataset_model.label}"')

        if not isinstance(irf, IrfMultiGaussian):
            raise ModelError(f'Irf in dataset "{dataset_model.label} is not a gaussian irf."')

        matrix_shape = (
            (global_axis.size, model_axis.size, self.order)
            if index_dependent(dataset_model)
            else (model_axis.size, self.order)
        )
        matrix = np.zeros(matrix_shape, dtype=np.float64)
        if index_dependent(dataset_model):
            centers, widths = [], []
            for global_index in range(global_axis.size):
                center, width = self.get_irf_parameter(irf, global_index, global_axis)
                centers.append(center)
                widths.append(width)
            _calculate_coherent_artifact_matrix(
                matrix,
                np.asarray(centers),
                np.asarray(widths),
                global_axis.size,
                model_axis,
                self.order,
            )

        else:
            center, width = self.get_irf_parameter(irf, None, global_axis)
            _calculate_coherent_artifact_matrix_on_index(
                matrix, center, width, model_axis, self.order
            )

        return self.compartments(), matrix

    def get_irf_parameter(
        self, irf: IrfMultiGaussian, global_index: int | None, global_axis: ArrayLike
    ) -> tuple[float, float]:
        center, width, _, shift, _, _ = irf.parameter(global_index, global_axis)
        center = center[0] - shift
        width = self.width.value if self.width is not None else width[0]
        return center, width

    def compartments(self):
        return [f"coherent_artifact_{i}" for i in range(1, self.order + 1)]

    def finalize_data(
        self,
        dataset_model: DatasetModel,
        dataset: xr.Dataset,
        is_full_model: bool = False,
        as_global: bool = False,
    ):
        global_dimension = dataset.attrs["global_dimension"]
        if not is_full_model:
            model_dimension = dataset.attrs["model_dimension"]
            dataset.coords["coherent_artifact_order"] = np.arange(1, self.order + 1)
            response_dimensions = (model_dimension, "coherent_artifact_order")
            if len(dataset.matrix.shape) == 3:
                response_dimensions = (global_dimension, *response_dimensions)
            dataset["coherent_artifact_response"] = (
                response_dimensions,
                dataset.matrix.sel(clp_label=self.compartments()).values,
            )
            dataset["coherent_artifact_associated_spectra"] = (
                (global_dimension, "coherent_artifact_order"),
                dataset.clp.sel(clp_label=self.compartments()).values,
            )
        retrieve_irf(dataset_model, dataset, global_dimension)


@nb.jit(nopython=True, parallel=False)
def _calculate_coherent_artifact_matrix(
    matrix, centers, widths, global_axis_size, model_axis, order
):
    for i in nb.prange(global_axis_size):
        _calculate_coherent_artifact_matrix_on_index(
            matrix[i], centers[i], widths[i], model_axis, order
        )


@nb.jit(nopython=True, parallel=True)
def _calculate_coherent_artifact_matrix_on_index(matrix, center, width, axis, order):
    matrix[:, 0] = np.exp(-1 * (axis - center) ** 2 / (2 * width**2))
    if order > 1:
        matrix[:, 1] = matrix[:, 0] * (center - axis) / width**2

    if order > 2:
        matrix[:, 2] = (
            matrix[:, 0] * (center**2 - width**2 - 2 * center * axis + axis**2) / width**4
        )



from __future__ import annotations

from dataclasses import dataclass

from glotaran.project.dataclass_helpers import asdict
from glotaran.project.dataclass_helpers import exclude_from_dict_field
from glotaran.project.dataclass_helpers import file_loadable_field
from glotaran.project.dataclass_helpers import fromdict
from glotaran.project.dataclass_helpers import init_file_loadable_fields


@dataclass
class DummyFileLoadable:
    def __init__(self, val: str) -> None:
        self.source_path = "dummy_file"
        self.data = {"foo": val}

    @classmethod
    def loader(
        cls: type[DummyFileLoadable],
        file_path: str,
    ) -> DummyFileLoadable:
        instance = cls(f"{file_path}_loaded")
        instance.source_path = file_path
        return instance


def test_serialize_to_file_name_field():
    @dataclass
    class DummyDataclass:
        foo: DummyFileLoadable = file_loadable_field(DummyFileLoadable)
        foo2: DummyFileLoadable = file_loadable_field(DummyFileLoadable)
        bar: int = exclude_from_dict_field(default=42)
        baz: int = 84

        def __post_init__(self):
            init_file_loadable_fields(self)

    dummy_class = DummyDataclass(foo=DummyFileLoadable.loader("foo.file"), foo2="foo2.file")

    assert dummy_class.foo.data == {"foo": "foo.file_loaded"}
    assert dummy_class.foo.source_path == "foo.file"
    assert dummy_class.foo2.data == {"foo": "foo2.file_loaded"}
    assert dummy_class.foo2.source_path == "foo2.file"
    assert dummy_class.bar == 42

    dummy_class_dict = asdict(dummy_class)

    assert dummy_class_dict["foo"] == "foo.file"
    assert dummy_class_dict["foo2"] == "foo2.file"
    assert "bar" not in dummy_class_dict

    assert dummy_class_dict["baz"] == 84
    assert dummy_class_dict["baz"] == dummy_class.baz

    loaded = fromdict(DummyDataclass, dummy_class_dict)

    assert loaded == dummy_class



"""Dummy module to test deprecate_module_attribute."""


def __getattr__(attribute_name: str):
    from glotaran.deprecation import deprecate_module_attribute

    if attribute_name == "deprecated_attribute":
        return deprecate_module_attribute(
            deprecated_qual_name=(
                "glotaran.deprecation.test.dummy_package.deprecated_module_attribute"
            ),
            new_qual_name="glotaran.deprecation.deprecation_utils.parse_version",
            to_be_removed_in_version="0.6.0",
        )

    if attribute_name == "foo_bar":
        return deprecate_module_attribute(
            deprecated_qual_name=("glotaran.deprecation.test.dummy_package.foo_bar"),
            new_qual_name="glotaran.does_not._need_to_exists",
            to_be_removed_in_version="0.6.0",
            module_load_overwrite="glotaran.deprecation.deprecation_utils.parse_version",
        )

    raise AttributeError(f"module {__name__} has no attribute {attribute_name}")



import pytest

from glotaran.io import DataIoInterface
from glotaran.io import ProjectIoInterface
from glotaran.model import Megacomplex
from glotaran.model import megacomplex
from glotaran.plugin_system.data_io_registration import known_data_formats
from glotaran.plugin_system.megacomplex_registration import known_megacomplex_names
from glotaran.plugin_system.project_io_registration import known_project_formats
from glotaran.testing.plugin_system import monkeypatch_plugin_registry
from glotaran.testing.plugin_system import monkeypatch_plugin_registry_data_io
from glotaran.testing.plugin_system import monkeypatch_plugin_registry_megacomplex
from glotaran.testing.plugin_system import monkeypatch_plugin_registry_project_io


@megacomplex()
class DummyMegacomplex(Megacomplex):
    pass


class DummyDataIo(DataIoInterface):
    pass


class DummyProjectIo(ProjectIoInterface):
    pass


def test_monkeypatch_megacomplexes():
    """Megacomplex only added to registry while context is entered."""
    with monkeypatch_plugin_registry_megacomplex(test_megacomplex={"test_mc": DummyMegacomplex}):
        assert "test_mc" in known_megacomplex_names()

    assert "test_mc" not in known_megacomplex_names()
    with monkeypatch_plugin_registry(test_megacomplex={"test_full": DummyMegacomplex}):
        assert "test_full" in known_megacomplex_names()

    assert "test_full" not in known_megacomplex_names()


def test_monkeypatch_data_io():
    """DataIoInterface only added to registry while context is entered."""
    with monkeypatch_plugin_registry_data_io(
        test_data_io={"test_dio": DummyDataIo(format_name="test")}
    ):
        assert "test_dio" in known_data_formats()

    assert "test_mc" not in known_data_formats()

    with monkeypatch_plugin_registry(test_data_io={"test_full": DummyDataIo(format_name="test")}):
        assert "test_full" in known_data_formats()

    assert "test_full" not in known_data_formats()


def test_monkeypatch_project_io():
    """ProjectIoInterface only added to registry while context is entered."""
    with monkeypatch_plugin_registry_project_io(
        test_project_io={"test_pio": DummyProjectIo(format_name="test")}
    ):
        assert "test_pio" in known_project_formats()

    assert "test_pio" not in known_megacomplex_names()
    with monkeypatch_plugin_registry(
        test_project_io={"test_full": DummyProjectIo(format_name="test")}
    ):
        assert "test_full" in known_project_formats()

    assert "test_full" not in known_project_formats()


@pytest.mark.parametrize("create_new_registry", (True, False))
def test_monkeypatch_plugin_registry_full(create_new_registry: bool):
    """Create a completely new registry."""

    assert "decay" in known_megacomplex_names()
    assert "yml" in known_project_formats()
    assert "sdt" in known_data_formats()

    with monkeypatch_plugin_registry(
        test_megacomplex={"test_mc": DummyMegacomplex},
        test_project_io={"test_pio": DummyProjectIo(format_name="test")},
        test_data_io={"test_dio": DummyDataIo(format_name="test")},
        create_new_registry=create_new_registry,
    ):
        assert "test_mc" in known_megacomplex_names()
        assert "test_pio" in known_project_formats()
        assert "test_dio" in known_data_formats()
        assert ("decay" not in known_megacomplex_names()) is create_new_registry
        assert ("yml" not in known_project_formats()) is create_new_registry
        assert ("sdt" not in known_data_formats()) is create_new_registry



"""Helper functions for attrs."""
from glotaran.utils.helpers import nan_or_equal


def no_default_vals_in_repr(cls):
    """Class decorator to omits attributes from repr that have their default value.

    Needs to be on top of the ``attr.define`` decorator.
    Based on: https://stackoverflow.com/a/47663099/3990615

    Parameters
    ----------
    cls
        Class decorated with ``attr.define``.

    Returns
    -------
    type[cls]
    """
    defaults = {
        attribute.name: attribute.default
        for attribute in cls.__attrs_attrs__
        if attribute.repr is True
    }

    def repr_(self) -> str:
        """Return string representing the instance.

        Parameters
        ----------
        self: cls
            Instance of the wrapped class.

        Returns
        -------
        str
        """
        real_cls = self.__class__

        if (qualname := getattr(real_cls, "__qualname__", None)) is not None:
            class_name = qualname.rsplit(">.", 1)[-1]
        else:
            class_name = real_cls.__name__

        args_str = ", ".join(
            f"{name}={repr(getattr(self, name))}"
            for name in defaults
            if not nan_or_equal(getattr(self, name), defaults[name])
        )

        return f"{class_name}({args_str})"

    cls.__repr__ = repr_
    return cls



"""Package with deprecation tests and helper functions."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import pytest

from glotaran.deprecation.deprecation_utils import GlotaranApiDeprecationWarning
from glotaran.deprecation.deprecation_utils import OverDueDeprecation

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Mapping
    from collections.abc import Sequence
    from typing import Any

    from _pytest.recwarn import WarningsRecorder


def deprecation_warning_on_call_test_helper(
    deprecated_callable: Callable[..., Any],
    *,
    raise_exception=False,
    args: Sequence[Any] = [],
    kwargs: Mapping[str, Any] = {},
) -> tuple[WarningsRecorder, Any]:
    """Helperfunction to quickly test that a deprecated class or function warns.

    By default this ignores error when calling the function/class,
    since those tests are only supposed to test the deprecation itself.

    However if the code is reimplemented it should be at least tested
    to return the proper type.


    Parameters
    ----------
    deprecated_callable : Callable[..., Any]
        Deprecated function or class.
    raise_exception : bool
        Whether or not to reraise an exception e.g. by calling with wrong args.
    args : Sequence[Any]
        Positional arguments for deprecated_callable.
    kwargs : Mapping[str, Any], optional
        Keyword arguments for deprecated_callable.

    Returns
    -------
    tuple[WarningsRecorder, Any]
        Tuple of the WarningsRecorder and return value of deprecated_callable

    Raises
    ------
    Exception
        Exception caused by deprecated_callable if raise_exception is True.
    """
    with pytest.warns(GlotaranApiDeprecationWarning) as record:
        try:
            result = deprecated_callable(*args, **kwargs)

            assert len(record) >= 1, f"{len(record)=}"
            assert Path(record[0].filename) == Path(
                __file__
            ), f"{Path(record[0].filename)=}, {Path(__file__)=}"

            return record, result

        except OverDueDeprecation as error:
            raise error

        except Exception as error:
            if raise_exception:
                raise error
            return record, None



"""Functions for data IO.

Note:
-----
Since Io functionality is purely plugin based this package mostly
reexports functions from the ``glotaran.plugin_system`` from a common place.
"""

from glotaran.io.interface import SAVING_OPTIONS_DEFAULT
from glotaran.io.interface import SAVING_OPTIONS_MINIMAL
from glotaran.io.interface import DataIoInterface
from glotaran.io.interface import ProjectIoInterface
from glotaran.io.interface import SavingOptions
from glotaran.io.prepare_dataset import prepare_time_trace_dataset
from glotaran.plugin_system.data_io_registration import data_io_plugin_table
from glotaran.plugin_system.data_io_registration import get_dataloader
from glotaran.plugin_system.data_io_registration import get_datasaver
from glotaran.plugin_system.data_io_registration import load_dataset
from glotaran.plugin_system.data_io_registration import register_data_io
from glotaran.plugin_system.data_io_registration import save_dataset
from glotaran.plugin_system.data_io_registration import set_data_plugin
from glotaran.plugin_system.data_io_registration import show_data_io_method_help
from glotaran.plugin_system.project_io_registration import get_project_io_method
from glotaran.plugin_system.project_io_registration import load_model
from glotaran.plugin_system.project_io_registration import load_parameters
from glotaran.plugin_system.project_io_registration import load_result
from glotaran.plugin_system.project_io_registration import load_scheme
from glotaran.plugin_system.project_io_registration import project_io_plugin_table
from glotaran.plugin_system.project_io_registration import register_project_io
from glotaran.plugin_system.project_io_registration import save_model
from glotaran.plugin_system.project_io_registration import save_parameters
from glotaran.plugin_system.project_io_registration import save_result
from glotaran.plugin_system.project_io_registration import save_scheme
from glotaran.plugin_system.project_io_registration import set_project_plugin
from glotaran.plugin_system.project_io_registration import show_project_io_method_help
from glotaran.utils.io import load_datasets



import numpy as np
import pytest

from glotaran.optimization.test.models import SimpleTestModel
from glotaran.parameter import Parameters
from glotaran.simulation import simulate


@pytest.mark.parametrize("index_dependent", [True, False])
@pytest.mark.parametrize("noise", [True, False])
def test_simulate_dataset(index_dependent, noise):
    model = SimpleTestModel(
        **{
            "megacomplex": {
                "m1": {"type": "simple-test-mc", "is_index_dependent": index_dependent},
                "m2": {"type": "simple-test-mc", "is_index_dependent": False},
            },
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                    "global_megacomplex": ["m2"],
                },
            },
        }
    )
    print(model.validate())
    assert model.valid()

    parameter = Parameters.from_list([1, 1])
    print(model.validate(parameter))
    assert model.valid(parameter)

    global_axis = np.asarray([1, 1, 1, 1])
    model_axis = np.asarray([2, 2, 2])

    data = simulate(
        model,
        "dataset1",
        parameter,
        {"global": global_axis, "model": model_axis},
        noise=noise,
        noise_std_dev=0.1,
    )
    assert np.array_equal(data["global"], global_axis)
    assert np.array_equal(data["model"], model_axis)
    assert data.data.shape == (3, 4)
    if not noise:
        assert np.array_equal(
            data.data,
            np.asarray(
                [
                    [2, 4, 6],
                    [4, 10, 16],
                    [6, 16, 26],
                    [8, 22, 36],
                ]
            ).T,
        )



from __future__ import annotations

import numpy as np

from glotaran.io import load_parameters


def test_parameter_group_copy():
    params = """
    a:
        - ["foo", 1, {non-negative: true, min: -1, max: 1, vary: false}]
        - 4
        - 5
    b:
        - 7
        - 8
    """
    parameters = load_parameters(params, format_name="yml_str")

    assert parameters.get("a.foo").value == 1
    assert parameters.get("a.foo").non_negative
    assert parameters.get("a.foo").minimum == -1
    assert parameters.get("a.foo").maximum == 1
    assert not parameters.get("a.foo").vary

    assert parameters.get("a.2").value == 4
    assert not parameters.get("a.2").non_negative
    assert parameters.get("a.2").minimum == -np.inf
    assert parameters.get("a.2").maximum == np.inf
    assert parameters.get("a.2").vary

    assert parameters.get("a.3").value == 5

    assert parameters.get("b.1").value == 7
    assert not parameters.get("b.1").non_negative
    assert parameters.get("b.1").minimum == -np.inf
    assert parameters.get("b.1").maximum == np.inf
    assert parameters.get("b.1").vary

    assert parameters.get("b.2").value == 8



"""Module containing a deprecated fake module."""
from glotaran.deprecation.deprecation_utils import deprecate_submodule

# just here to be tested
deprecated_module = deprecate_submodule(
    deprecated_module_name="glotaran.deprecation.test.dummy_package.deprecated_module",
    new_module_name="glotaran.deprecation.deprecation_utils",
    to_be_removed_in_version="0.6.0",
)
overwritten_module = deprecate_submodule(
    deprecated_module_name="glotaran.deprecation.test.dummy_package.overwritten_module",
    new_module_name="glotaran.does_not._need_to_exists",
    to_be_removed_in_version="0.6.0",
    module_load_overwrite="glotaran.deprecation.deprecation_utils",
)



"""Tests for ``glotaran.utils.numeric_helpers``."""

from typing import Any

import numpy as np
import pytest

from glotaran.utils.helpers import nan_or_equal


@pytest.mark.parametrize(
    "lhs, rhs, expected",
    (
        ("foo", "foo", True),
        (np.nan, np.nan, True),
        (1, 1, True),
        (1, 2, False),
        ("foo", "bar", False),
    ),
)
def test_nan_or_equal(lhs: Any, rhs: Any, expected: bool):
    """Only ``False`` if values actually differ."""
    assert nan_or_equal(lhs, rhs) == expected



"""Module containing CSV io support."""

from __future__ import annotations

import numpy as np
import pandas as pd

from glotaran.io import ProjectIoInterface
from glotaran.io import register_project_io
from glotaran.parameter import Parameters
from glotaran.parameter.parameter import OPTION_NAMES_DESERIALIZED
from glotaran.utils.io import safe_dataframe_fillna
from glotaran.utils.io import safe_dataframe_replace


@register_project_io(["csv"])
class CsvProjectIo(ProjectIoInterface):
    """Plugin for CSV data io."""

    def load_parameters(self, file_name: str, sep: str = ",") -> Parameters:
        """Load parameters from CSV file.

        Parameters
        ----------
        file_name : str
            Name of file to be loaded.
        sep: str
            Other separators can be used optionally., by default ','

        Returns
        -------
            :class:`Parameters
        """
        df = pd.read_csv(file_name, skipinitialspace=True, na_values=["None", "none"], sep=sep)
        df.columns = [column.lower() for column in df.columns]
        df.rename(columns=OPTION_NAMES_DESERIALIZED, inplace=True)
        safe_dataframe_fillna(df, "minimum", -np.inf)
        safe_dataframe_fillna(df, "maximum", np.inf)
        return Parameters.from_dataframe(df, source=file_name)

    def save_parameters(
        self,
        parameters: Parameters,
        file_name: str,
        *,
        sep: str = ",",
        as_optimized: bool = True,
        replace_infinfinity: bool = True,
    ) -> None:
        """Save a :class:`Parameters` to a CSV file.

        Parameters
        ----------
        parameters : Parameters
            Parameters to be saved to file.
        file_name : str
            File to write the parameters to.
        sep: str
            Other separators can be used optionally., by default ','
        as_optimized : bool
            Weather to include properties which are the result of optimization.
        replace_infinfinity : bool
            Weather to replace infinity values with empty strings.
        """
        df = parameters.to_dataframe()
        if replace_infinfinity is True:
            safe_dataframe_replace(df, "minimum", -np.inf, "")
            safe_dataframe_replace(df, "maximum", np.inf, "")
        df.to_csv(file_name, na_rep="None", index=False, sep=sep)



from __future__ import annotations

import sys
from collections.abc import Iterable

import click
from click import echo
from click import prompt

from glotaran.io import ProjectIoInterface
from glotaran.io import load_dataset
from glotaran.io import load_model
from glotaran.io import load_parameters
from glotaran.io import load_scheme
from glotaran.plugin_system.base_registry import methods_differ_from_baseclass_table
from glotaran.plugin_system.project_io_registration import get_project_io
from glotaran.plugin_system.project_io_registration import known_project_formats


def signature_analysis(cmd):
    cmd = click.option(
        "--model_file",
        "-m",
        default=None,
        type=click.Path(exists=True, dir_okay=False),
        help="Path to model file.",
    )(cmd)
    cmd = click.option(
        "--parameters_file",
        "-p",
        default=None,
        type=click.Path(exists=True, dir_okay=False),
        help="(optional) Path to parameter file.",
    )(cmd)
    cmd = click.argument(
        "scheme_file", type=click.Path(exists=True, dir_okay=False), required=False
    )(cmd)
    return cmd


def _load_file(filename, loader, name, verbose):
    try:
        if verbose:
            echo(f"Loading {name} file '{filename}'... ", nl=False)
        result = loader(filename)
        if verbose:
            echo("Success")
        return result
    except Exception as e:
        if verbose:
            echo(message="Error", err=True)
        else:
            echo(message=f"Error parsing {name} file: \n", err=True)
        echo(message=e, err=True)
        sys.exit(1)


def load_scheme_file(filename, verbose=False):
    return _load_file(
        filename, lambda file: load_scheme(file, format_name="yml"), "scheme", verbose
    )


def load_model_file(filename, verbose=False):
    return _load_file(filename, lambda file: load_model(file, format_name="yml"), "model", verbose)


def load_parameter_file(filename, fmt=None, verbose=False):
    def loader(filename):
        return load_parameters(filename, format_name=fmt)

    return _load_file(filename, loader, "parameter", verbose)


def load_dataset_file(filename, fmt=None, verbose=False):
    def loader(filename):
        return load_dataset(filename, format_name=fmt)

    return _load_file(filename, loader, "parameter", verbose)


def select_name(filename, dataset):
    names = list(dataset)
    echo(f"\nDataset names in in '{filename}':\n")
    for i, n in enumerate(names):
        echo(f"* [{i}] {n}")

    n = prompt(
        "\n Please select a name to export",
        type=click.IntRange(min=0, max=len(names) - 1),
        show_choices=True,
    )
    return names[n]


def select_data(data, dim, selection):
    try:
        [float(i) for i in data.coords[dim].values]
        numeric = True
    except ValueError:
        numeric = False
    method = None
    if numeric:
        try:
            if isinstance(selection, tuple):
                selection = tuple(float(c) for c in selection)
            elif isinstance(selection, list):
                selection = [float(c) for c in selection]
            else:
                selection = float(selection)
        except ValueError:
            raise ValueError(f"Error: Selection '{selection}' is not numeric")
        method = "nearest"
    if isinstance(selection, tuple):
        min = selection[0]
        max = selection[1]
        selection = data.coords[dim]
        selection = selection.where(selection >= min, drop=True)
        selection = selection.where(selection <= max, drop=True)
    return data.sel({dim: selection}, method=method)


def write_data(data, out):
    df = data.to_dataframe()
    if len(data.dims) == 2:
        df = df.reset_index().pivot(index=data.dims[0], columns=data.dims[1], values=data.name)
    df.to_csv(out)


def project_io_list_supporting_plugins(
    method_name: str, block_list: Iterable[str] | None = None
) -> Iterable[str]:
    """List all project-io plugin that implement ``method_name``.

    Parameters
    ----------
    method_name: str
        Name of the method which should be supported.
    block_list: Iterable[str]
        Iterable of plugin names which should be omitted.
    """
    if block_list is None:
        block_list = []
    support_table = methods_differ_from_baseclass_table(
        method_names=method_name,
        plugin_registry_keys=known_project_formats(full_names=False),
        get_plugin_function=get_project_io,
        base_class=ProjectIoInterface,
    )
    support_table = filter(lambda entry: entry[1], support_table)
    supporting_list: Iterable[str] = (entry[0].replace("`", "") for entry in support_table)
    return list(filter(lambda entry: entry not in block_list, supporting_list))


class ValOrRangeOrList(click.ParamType):
    name = "number or range or list"

    def convert(self, value, param, ctx):
        if value[0] == "(" and value[-1] == ")":
            split = value[1:-1].split(",")
            if len(split) != 2:
                self.fail(f"Malformed range: '{value}'")
            return (split[0].strip(), split[1].strip())
        if value[0] == "[" and value[-1] == "]":
            split = value[1:-1].split(",")
            return [s.strip() for s in split]
        return value


VALORRANGEORLIST = ValOrRangeOrList()



import subprocess
from pathlib import Path

from click.testing import CliRunner

from glotaran.cli import main


def test_cli_deprecation():
    """Needs to be tested with subprocess since CliRunner clears stdout."""

    result = subprocess.run("glotaran", capture_output=True, check=True)

    assert result.stderr.decode().startswith(
        "[DEPRECATED] The pyglotaran command line interface will be removed without"
    )


def test_cli_help():
    """Test the CLI help options."""
    runner = CliRunner()
    result = runner.invoke(main)
    assert result.exit_code == 0
    help_result = runner.invoke(main, ["--help"], prog_name="glotaran")
    assert help_result.exit_code == 0
    assert "Usage: glotaran [OPTIONS] COMMAND [ARGS]..." in help_result.output


def test_cli_pluginlist():
    """Test the CLI pluginlist option."""
    runner = CliRunner()
    result = runner.invoke(main, ["pluginlist"], prog_name="glotaran")
    assert result.exit_code == 0
    assert "Installed Glotaran Plugins" in result.output


def test_cli_validate_parameters_file(tmp_path: Path):
    """Test the CLI pluginlist option."""
    empty_file = tmp_path.joinpath("empty_file.yml")
    empty_file.touch()
    runner = CliRunner()
    result_ok = runner.invoke(
        main, ["validate", "--parameters_file", str(empty_file)], prog_name="glotaran"
    )
    assert result_ok.exit_code == 0
    assert "Type 'glotaran validate --help' for more info." in result_ok.output
    non_existing_file = tmp_path.joinpath("_does_not_exist_.yml")
    result_file_not_exist = runner.invoke(
        main, ["validate", "--parameters_file", str(non_existing_file)], prog_name="glotaran"
    )
    assert result_file_not_exist.exit_code == 2
    assert all(
        substring in result_file_not_exist.output for substring in ("Error", "does not exist")
    )



"""Module containing the YAML Data and Project IO plugins."""
from __future__ import annotations

from dataclasses import replace
from pathlib import Path
from typing import TYPE_CHECKING

from glotaran.builtin.io.yml.utils import load_dict
from glotaran.builtin.io.yml.utils import write_dict
from glotaran.deprecation.modules.builtin_io_yml import model_spec_deprecations
from glotaran.io import SAVING_OPTIONS_DEFAULT
from glotaran.io import ProjectIoInterface
from glotaran.io import SavingOptions
from glotaran.io import register_project_io
from glotaran.io import save_model
from glotaran.io import save_result
from glotaran.io import save_scheme
from glotaran.model import Model
from glotaran.parameter import Parameters
from glotaran.plugin_system.megacomplex_registration import get_megacomplex
from glotaran.project.dataclass_helpers import asdict
from glotaran.project.dataclass_helpers import fromdict
from glotaran.project.project import Result
from glotaran.project.scheme import Scheme
from glotaran.utils.sanitize import sanitize_yaml

if TYPE_CHECKING:
    from typing import Any


@register_project_io(["yml", "yaml", "yml_str"])
class YmlProjectIo(ProjectIoInterface):
    """Plugin for YAML project io."""

    def load_model(self, file_name: str) -> Model:
        """Load a :class:`Model` from a model specification in a yaml file.

        Parameters
        ----------
        file_name: str
            Path to the model file to read.

        Raises
        ------
        ValueError
            If ``megacomplex`` was not provided in the model specification.
        ValueError
            If ``default_megacomplex`` was not provided and any megacomplex is missing the type
            attribute.

        Returns
        -------
        Model
        """
        spec = self._load_yml(file_name)

        model_spec_deprecations(spec)

        spec = sanitize_yaml(spec)

        if "megacomplex" not in spec:
            raise ValueError("No megacomplex defined in model")

        default_megacomplex = spec.pop("default_megacomplex", None)

        if default_megacomplex is None and any(
            "type" not in m for m in spec["megacomplex"].values()
        ):
            raise ValueError(
                "Default megacomplex is not defined in model and "
                "at least one megacomplex does not have a type."
            )

        spec["megacomplex"] = {
            label: m | {"type": default_megacomplex} if "type" not in m else m
            for label, m in spec["megacomplex"].items()
        }

        megacomplex_types = {get_megacomplex(m["type"]) for m in spec["megacomplex"].values()}
        return Model.create_class_from_megacomplexes(megacomplex_types)(**spec)

    def save_model(self, model: Model, file_name: str):
        """Save a :class:`Model` instance to a specification file.

        Parameters
        ----------
        model: Model
            Model instance to save to specs file.
        file_name : str
            File to write the model specs to.
        """
        model_dict = model.as_dict()
        # We replace tuples with strings
        for items in model_dict.values():
            if not isinstance(items, (list, dict)):
                continue
            item_iterator = items if isinstance(items, list) else items.values()
            for item in item_iterator:
                for prop_name, prop in item.items():
                    if isinstance(prop, dict) and any(isinstance(k, tuple) for k in prop):
                        keys = [f"({k[0]}, {k[1]})" for k in prop]
                        item[prop_name] = {f"{k}": v for k, v in zip(keys, prop.values())}
        write_dict(model_dict, file_name=file_name)

    def load_parameters(self, file_name: str) -> Parameters:
        """Load :class:`Parameters` instance from the specification defined in ``file_name``.

        Parameters
        ----------
        file_name: str
            File containing the parameter specification.

        Returns
        -------
        Parameters
        """  # noqa:  D414
        spec = self._load_yml(file_name)

        if isinstance(spec, list):
            return Parameters.from_list(spec)
        else:
            return Parameters.from_dict(spec)

    def load_scheme(self, file_name: str) -> Scheme:
        """Load :class:`Scheme` instance from the specification defined in ``file_name``.

        Parameters
        ----------
        file_name: str
            File containing the scheme specification.

        Returns
        -------
        Scheme
        """
        spec = self._load_yml(file_name)
        return fromdict(Scheme, spec, folder=Path(file_name).parent)

    def save_scheme(self, scheme: Scheme, file_name: str):
        """Write a :class:`Scheme` instance to a specification file ``file_name``.

        Parameters
        ----------
        scheme: Scheme
            :class:`Scheme` instance to save to file.
        file_name: str
            Path to the file to write the scheme specification to.
        """
        scheme_dict = asdict(scheme, folder=Path(file_name).parent)
        write_dict(scheme_dict, file_name=file_name)

    def load_result(self, result_path: str) -> Result:
        """Create a :class:`Result` instance from the specs defined in a file.

        Parameters
        ----------
        result_path : str
            Path containing the result data.

        Returns
        -------
        Result
            :class:`Result` instance created from the saved format.
        """
        result_file_path = Path(result_path)
        if result_file_path.suffix not in [".yml", ".yaml"]:
            result_file_path = result_file_path / "result.yml"
        spec = self._load_yml(result_file_path.as_posix())
        if "number_of_data_points" in spec:
            spec["number_of_residuals"] = spec.pop("number_of_data_points")
        if "number_of_parameters" in spec:
            spec["number_of_free_parameters"] = spec.pop("number_of_parameters")
        return fromdict(Result, spec, folder=result_file_path.parent)

    def save_result(
        self,
        result: Result,
        result_path: str,
        saving_options: SavingOptions = SAVING_OPTIONS_DEFAULT,
    ) -> list[str]:
        """Write a :class:`Result` instance to a specification file and data files.

        Returns a list with paths of all saved items.
        The following files are saved if not configured otherwise:
        * ``result.md``: The result with the model formatted as markdown text.
        * ``result.yml``: Yaml spec file of the result
        * ``model.yml``: Model spec file.
        * ``scheme.yml``: Scheme spec file.
        * ``initial_parameters.csv``: Initially used parameters.
        * ``optimized_parameters.csv``: The optimized parameter as csv file.
        * ``parameter_history.csv``: Parameter changes over the optimization
        * ``optimization_history.csv``: Parsed table printed by the SciPy optimizer
        * ``{dataset_label}.nc``: The result data for each dataset as NetCDF file.

        Parameters
        ----------
        result: Result
            :class:`Result` instance to write.
        result_path: str
            Path to write the result data to.
        saving_options: SavingOptions
            Options for saving the the result.

        Returns
        -------
        list[str]
            List of file paths which were created.
        """
        result_file_path = Path(result_path)
        if result_file_path.suffix not in [".yml", ".yaml"]:
            result_file_path = result_file_path / "result.yml"
        result_folder = result_file_path.parent
        paths = save_result(
            result,
            result_folder,
            format_name="folder",
            saving_options=saving_options,
            allow_overwrite=True,
            used_inside_of_plugin=True,
        )

        model_path = result_folder / "model.yml"
        save_model(result.scheme.model, model_path, allow_overwrite=True)
        paths.append(model_path.as_posix())

        # The source_path attribute of the datasets only gets changed for `result.data`
        # Which why we overwrite the data attribute on a copy of `result.scheme`
        scheme = replace(result.scheme, data=result.data)
        scheme_path = result_folder / "scheme.yml"
        save_scheme(scheme, scheme_path, allow_overwrite=True)
        paths.append(scheme_path.as_posix())

        result_dict = asdict(result, folder=result_folder)
        write_dict(result_dict, file_name=result_file_path)
        paths.append(result_file_path.as_posix())

        return paths

    def _load_yml(self, file_name: str) -> dict[str, Any]:
        return load_dict(file_name, self.format != "yml_str")



"""Mock functionality for the plugin system."""
from __future__ import annotations

from contextlib import ExitStack
from contextlib import contextmanager
from typing import TYPE_CHECKING
from unittest import mock

from glotaran.plugin_system.base_registry import __PluginRegistry

if TYPE_CHECKING:
    from collections.abc import Generator
    from collections.abc import MutableMapping

    from glotaran.io.interface import DataIoInterface
    from glotaran.io.interface import ProjectIoInterface
    from glotaran.model.megacomplex import Megacomplex
    from glotaran.plugin_system.base_registry import _PluginType


@contextmanager
def _monkeypatch_plugin_registry(
    register_name: str,
    test_registry: MutableMapping[str, _PluginType] | None = None,
    create_new_registry: bool = False,
) -> Generator[None, None, None]:
    """Contextmanager to monkeypatch any Pluginregistry with name ``register_name``.

    Parameters
    ----------
    register_name : str
        Name of the register which should be patched.
    test_registry : MutableMapping[str, _PluginType]
        Registry to to update or replace the ``register_name`` registry with.
        , by default None
    create_new_registry : bool
        Whether to update the actual registry or create a new one from ``test_registry``
        , by default False

    Yields
    ------
    Generator[None, None, None]
        Just to keep the context alive.

    See Also
    --------
    monkeypatch_plugin_registry_megacomplex
    monkeypatch_plugin_registry_data_io
    monkeypatch_plugin_registry_project_io
    """
    if test_registry is not None:
        initial_plugins = {} if create_new_registry else __PluginRegistry.__dict__[register_name]

        with mock.patch.object(
            __PluginRegistry, register_name, {**initial_plugins, **test_registry}
        ):
            yield
    else:
        yield


@contextmanager
def monkeypatch_plugin_registry_megacomplex(
    test_megacomplex: MutableMapping[str, type[Megacomplex]] | None = None,
    create_new_registry: bool = False,
) -> Generator[None, None, None]:
    """Monkeypatch the :class:`Megacomplex` registry.

    Parameters
    ----------
    test_megacomplex : MutableMapping[str, type[Megacomplex]], optional
        Registry to to update or replace the ``Megacomplex`` registry with.
        , by default None
    create_new_registry : bool
        Whether to update the actual registry or create a new one from ``test_megacomplex``
        , by default False

    Yields
    ------
    Generator[None, None, None]
        Just to keep the context alive.
    """
    with _monkeypatch_plugin_registry("megacomplex", test_megacomplex, create_new_registry):
        yield


@contextmanager
def monkeypatch_plugin_registry_data_io(
    test_data_io: MutableMapping[str, DataIoInterface] | None = None,
    create_new_registry: bool = False,
) -> Generator[None, None, None]:
    """Monkeypatch the :class:`DataIoInterface` registry.

    Parameters
    ----------
    test_data_io : MutableMapping[str, DataIoInterface], optional
        Registry to to update or replace the ``DataIoInterface`` registry with.
        , by default None
    create_new_registry : bool
        Whether to update the actual registry or create a new one from ``test_data_io``
        , by default False

    Yields
    ------
    Generator[None, None, None]
        Just to keep the context alive.
    """
    with _monkeypatch_plugin_registry("data_io", test_data_io, create_new_registry):
        yield


@contextmanager
def monkeypatch_plugin_registry_project_io(
    test_project_io: MutableMapping[str, ProjectIoInterface] | None = None,
    create_new_registry: bool = False,
) -> Generator[None, None, None]:
    """Monkeypatch the :class:`ProjectIoInterface` registry.

    Parameters
    ----------
    test_project_io : MutableMapping[str, ProjectIoInterface], optional
        Registry to to update or replace the ``ProjectIoInterface`` registry with.
        , by default None
    create_new_registry : bool
        Whether to update the actual registry or create a new one from ``test_data_io``
        , by default False

    Yields
    ------
    Generator[None, None, None]
        Just to keep the context alive.
    """
    with _monkeypatch_plugin_registry("project_io", test_project_io, create_new_registry):
        yield


@contextmanager
def monkeypatch_plugin_registry(
    *,
    test_megacomplex: MutableMapping[str, type[Megacomplex]] | None = None,
    test_data_io: MutableMapping[str, DataIoInterface] | None = None,
    test_project_io: MutableMapping[str, ProjectIoInterface] | None = None,
    create_new_registry: bool = False,
) -> Generator[None, None, None]:
    """Contextmanager to monkeypatch multiple plugin registries at once.

    Parameters
    ----------
    test_megacomplex : MutableMapping[str, type[Megacomplex]], optional
        Registry to to update or replace the ``Megacomplex`` registry with.
        , by default None
    test_data_io : MutableMapping[str, DataIoInterface], optional
        Registry to to update or replace the ``DataIoInterface`` registry with.
        , by default None
    test_project_io : MutableMapping[str, ProjectIoInterface], optional
        Registry to to update or replace the ``ProjectIoInterface`` registry with.
        , by default None
    create_new_registry : bool
        Whether to update the actual registry or create a new one from the arguments.
        , by default False

    Yields
    ------
    Generator[None, None, None]
        Just keeps all context manager alive

    See Also
    --------
    monkeypatch_plugin_registry_megacomplex
    monkeypatch_plugin_registry_data_io
    monkeypatch_plugin_registry_project_io
    """
    context_managers = [
        monkeypatch_plugin_registry_megacomplex(test_megacomplex, create_new_registry),
        monkeypatch_plugin_registry_data_io(test_data_io, create_new_registry),
        monkeypatch_plugin_registry_project_io(test_project_io, create_new_registry),
    ]

    with ExitStack() as stack:
        for context_manager in context_managers:
            stack.enter_context(context_manager)
        yield



import pytest

from glotaran.builtin.megacomplexes.decay import DecayParallelMegacomplex
from glotaran.builtin.megacomplexes.decay import DecaySequentialMegacomplex
from glotaran.builtin.megacomplexes.spectral import SpectralMegacomplex
from glotaran.project.generators.generator import generate_model


@pytest.mark.parametrize("megacomplex_type", ["parallel", "sequential"])
@pytest.mark.parametrize("irf", [True, False])
@pytest.mark.parametrize("spectral", [True, False])
def test_generate_parallel_model(megacomplex_type: str, irf: bool, spectral: bool):
    nr_compartments = 5
    expected_compartments = [f"species_{i+1}" for i in range(nr_compartments)]
    model_type = f"spectral_decay_{megacomplex_type}" if spectral else f"decay_{megacomplex_type}"
    model = generate_model(
        generator_name=model_type,
        generator_arguments={
            "nr_compartments": nr_compartments,
            "irf": irf,
        },
    )
    print(model)

    assert f"megacomplex_{megacomplex_type}_decay" in model.megacomplex
    megacomplex = model.megacomplex[f"megacomplex_{megacomplex_type}_decay"]
    assert isinstance(megacomplex, (DecayParallelMegacomplex, DecaySequentialMegacomplex))
    assert megacomplex.type == f"decay-{megacomplex_type}"
    assert megacomplex.compartments == expected_compartments
    assert megacomplex.rates == [f"rates.species_{i+1}" for i in range(nr_compartments)]

    assert "dataset_1" in model.dataset
    dataset = model.dataset["dataset_1"]
    assert dataset.megacomplex == [f"megacomplex_{megacomplex_type}_decay"]

    if spectral:
        assert "megacomplex_spectral" in model.megacomplex
        megacomplex = model.megacomplex["megacomplex_spectral"]
        assert isinstance(megacomplex, SpectralMegacomplex)
        assert expected_compartments == list(megacomplex.shape.keys())
        expected_shapes = [f"shape_species_{i+1}" for i in range(nr_compartments)]
        assert expected_shapes == list(megacomplex.shape.values())

        for i, shape in enumerate(expected_shapes):
            assert shape in model.shape  # type:ignore[attr-defined]
            assert model.shape[shape].type == "gaussian"  # type:ignore[attr-defined]
            assert (
                model.shape[shape].amplitude  # type:ignore[attr-defined]
                == f"shapes.species_{i+1}.amplitude"
            )
            assert (
                model.shape[shape].location  # type:ignore[attr-defined]
                == f"shapes.species_{i+1}.location"
            )
            assert (
                model.shape[shape].width  # type:ignore[attr-defined]
                == f"shapes.species_{i+1}.width"
            )
            assert dataset.global_megacomplex == ["megacomplex_spectral"]

    if irf:
        assert dataset.irf == "gaussian_irf"  # type:ignore[attr-defined]
        assert "gaussian_irf" in model.irf  # type:ignore[attr-defined]
        assert (
            model.irf["gaussian_irf"].center  # type:ignore[attr-defined]
            == "irf.center"
        )
        assert model.irf["gaussian_irf"].width == "irf.width"  # type:ignore[attr-defined]



"""Tests for glotaran.model.dataset_model.DatasetModel"""
from __future__ import annotations

import pytest

from glotaran.model.dataset_model import get_dataset_model_model_dimension
from glotaran.model.item import fill_item
from glotaran.model.item import get_item_issues
from glotaran.model.megacomplex import Megacomplex
from glotaran.model.megacomplex import megacomplex
from glotaran.model.model import Model
from glotaran.parameter import Parameters


@megacomplex()
class MockMegacomplexNonUniqueExclusive(Megacomplex):
    type: str = "test_megacomplex_not_exclusive_unique"


@megacomplex(exclusive=True)
class MockMegacomplexExclusive(Megacomplex):
    type: str = "test_megacomplex_exclusive"


@megacomplex(unique=True)
class MockMegacomplexUnique(Megacomplex):
    type: str = "test_megacomplex_unique"


@megacomplex()
class MockMegacomplexDim1(Megacomplex):
    dimension: str = "dim1"
    type: str = "test_megacomplex_dim1"


@megacomplex()
class MockMegacomplexDim2(Megacomplex):
    dimension: str = "dim2"
    type: str = "test_megacomplex_dim2"


def test_get_issues_datasetmodel():
    mcls = Model.create_class_from_megacomplexes(
        [
            MockMegacomplexNonUniqueExclusive,
            MockMegacomplexExclusive,
            MockMegacomplexUnique,
        ]
    )
    m = mcls(
        megacomplex={
            "m": {"type": "test_megacomplex_not_exclusive_unique"},
            "m_exclusive": {"type": "test_megacomplex_exclusive"},
            "m_unique": {"type": "test_megacomplex_unique"},
        },
        dataset={
            "ok": {"megacomplex": ["m"]},
            "exclusive": {"megacomplex": ["m", "m_exclusive"]},
            "unique": {"megacomplex": ["m_unique", "m_unique"]},
        },
    )

    assert len(get_item_issues(item=m.dataset["ok"], model=m)) == 0
    assert len(get_item_issues(item=m.dataset["exclusive"], model=m)) == 1
    assert len(get_item_issues(item=m.dataset["unique"], model=m)) == 2

    m = mcls(
        megacomplex={
            "m": {"type": "test_megacomplex_not_exclusive_unique"},
            "m_exclusive": {"type": "test_megacomplex_exclusive"},
            "m_unique": {"type": "test_megacomplex_unique"},
        },
        dataset={
            "ok": {"megacomplex": [], "global_megacomplex": ["m"]},
            "exclusive": {"megacomplex": [], "global_megacomplex": ["m", "m_exclusive"]},
            "unique": {"megacomplex": [], "global_megacomplex": ["m_unique", "m_unique"]},
        },
    )

    assert len(get_item_issues(item=m.dataset["ok"], model=m)) == 0
    assert len(get_item_issues(item=m.dataset["exclusive"], model=m)) == 1
    assert len(get_item_issues(item=m.dataset["unique"], model=m)) == 2


def test_get_model_dim():
    mcls = Model.create_class_from_megacomplexes([MockMegacomplexDim1, MockMegacomplexDim2])
    m = mcls(
        megacomplex={
            "m1": {"type": "test_megacomplex_dim1"},
            "m2": {"type": "test_megacomplex_dim2"},
        },
        dataset={
            "ok": {"megacomplex": ["m1"]},
            "error1": {"megacomplex": []},
            "error2": {"megacomplex": ["m1", "m2"]},
        },
    )

    get_dataset_model_model_dimension(
        fill_item(m.dataset["ok"], model=m, parameters=Parameters.from_list([]))
    )
    with pytest.raises(ValueError, match="Dataset model 'ok' was not filled."):
        get_dataset_model_model_dimension(m.dataset["ok"])
    with pytest.raises(ValueError, match="No megacomplex set for dataset model 'error1'."):
        get_dataset_model_model_dimension(m.dataset["error1"])
    with pytest.raises(
        ValueError, match="Megacomplex dimensions do not match for dataset model 'error2'."
    ):
        get_dataset_model_model_dimension(
            fill_item(m.dataset["error2"], model=m, parameters=Parameters.from_list([]))
        )



"""This module contains the model."""
from __future__ import annotations

from collections.abc import Callable
from collections.abc import Generator
from collections.abc import Iterable
from collections.abc import Mapping
from typing import Any
from typing import ClassVar
from uuid import uuid4

from attr import asdict
from attr import fields
from attr import ib
from attrs import Attribute
from attrs import define
from attrs import make_class
from attrs import resolve_types

from glotaran.io import load_model
from glotaran.model.clp_constraint import ClpConstraint
from glotaran.model.clp_penalties import ClpPenalty
from glotaran.model.clp_relation import ClpRelation
from glotaran.model.dataset_group import DatasetGroup
from glotaran.model.dataset_group import DatasetGroupModel
from glotaran.model.dataset_model import DatasetModel
from glotaran.model.item import Item
from glotaran.model.item import ItemIssue
from glotaran.model.item import ModelItem
from glotaran.model.item import TypedItem
from glotaran.model.item import get_item_issues
from glotaran.model.item import item_to_markdown
from glotaran.model.item import iterate_parameter_names_and_labels
from glotaran.model.item import model_attributes
from glotaran.model.item import strip_type_and_structure_from_attribute
from glotaran.model.megacomplex import Megacomplex
from glotaran.model.weight import Weight
from glotaran.parameter import Parameter
from glotaran.parameter import Parameters
from glotaran.utils.ipython import MarkdownStr

DEFAULT_DATASET_GROUP = "default"
META_ITEMS = "__glotaran_items__"
META = {META_ITEMS: True}


class ModelError(Exception):
    """Raised when a model contains errors."""

    def __init__(self, error: str):
        """Create a model error.

        Parameters
        ----------
        error: str
            The error string.
        """
        super().__init__(f"ModelError: {error}")


def _load_item_from_dict(
    item_type: type[Item], value: Item | Mapping, extra: dict[str, Any] | None = None
) -> Item:
    """Load an item from a dictionary.

    Parameters
    ----------
    item_type: type[Item]
        The item type.
    value: Item | dict
        The value to load from.
    extra: dict[str, Any] | None
        Extra arguments for the item.

    Returns
    -------
    Item

    Raises
    ------
    ModelError
        Raised if a modelitem is missing.
    """
    if not isinstance(value, Item):
        if extra:
            value = value | extra
        if issubclass(item_type, TypedItem):
            try:
                item_type = item_type.get_item_type_class(value["type"])
            except KeyError:
                raise ModelError(f"Missing 'type' for item {item_type}")
        return item_type(**(value))
    return value


def _load_model_items_from_dict(
    item_type: type[Item], item_dict: Mapping[str, ModelItem | dict]
) -> dict[str, ModelItem]:
    """Load a model items from a dictionary.

    Parameters
    ----------
    item_type: type[Item]
        The item type.
    item_dict: dict[str, ModelItem | dict]
        The item dictionary.

    Returns
    -------
    dict[str, ModelItem]
    """
    return {
        label: _load_item_from_dict(item_type, value, extra={"label": label})  # type:ignore[misc]
        for label, value in item_dict.items()
    }


def _load_global_items_from_dict(
    item_type: type[Item], item_list: list[Item | dict]
) -> list[Item]:
    """Load an item from a dictionary.

    Parameters
    ----------
    item_type: type[Item]
        The item type.
    item_list: list[Item | dict]
        The list of item dicts.

    Returns
    -------
    list[Item]
    """
    return [_load_item_from_dict(item_type, value) for value in item_list]


def _load_dataset_groups(
    dataset_groups: dict[str, DatasetGroupModel | Any]
) -> dict[str, DatasetGroupModel]:
    """Add the default dataset group if not present.

    Parameters
    ----------
    dataset_groups: dict[str, DatasetGroupModel]
        The dataset groups.

    Returns
    -------
    dict[str, DatasetGroupModel]
    """
    dataset_group_items = _load_model_items_from_dict(DatasetGroupModel, dataset_groups)
    if DEFAULT_DATASET_GROUP not in dataset_group_items:
        dataset_group_items[DEFAULT_DATASET_GROUP] = DatasetGroupModel(  # type:ignore[call-arg]
            label=DEFAULT_DATASET_GROUP
        )
    return dataset_group_items  # type:ignore[return-value]


def _global_item_attribute(item_type: type[Item]) -> Attribute:
    """Create a global item attribute.

    Parameters
    ----------
    item_type: type[Item]
        The item type.

    Returns
    -------
    Attribute
    """
    return ib(
        factory=list,
        converter=lambda value: _load_global_items_from_dict(item_type, value),
        metadata=META,
    )


def _model_item_attribute(item_type: type[ModelItem]):
    """Create a model item attribute.

    Parameters
    ----------
    item_type: type[ModelItem]
        The item type.

    Returns
    -------
    Attribute
    """
    return ib(
        type=dict[str, item_type],  # type:ignore[valid-type]
        factory=dict,
        converter=lambda value: _load_model_items_from_dict(item_type, value),
        metadata=META,
    )


def _create_attributes_for_item(item_type: type[Item]) -> dict[str, Attribute]:
    """Create attributes for an item.

    Parameters
    ----------
    item_type: type[Item]
        The item type.

    Returns
    -------
    dict[str, Attribute]
    """
    attributes = {}
    for model_item in model_attributes(item_type, with_alias=False):
        _, model_item_type = strip_type_and_structure_from_attribute(model_item)
        attributes[model_item.name] = _model_item_attribute(model_item_type)
    return attributes


@define(kw_only=True)
class Model:
    """A model for global target analysis."""

    loader: ClassVar[Callable] = load_model

    source_path: str | None = ib(default=None, init=False, repr=False)
    clp_penalties: list[ClpPenalty] = _global_item_attribute(ClpPenalty)
    clp_constraints: list[ClpConstraint] = _global_item_attribute(ClpConstraint)
    clp_relations: list[ClpRelation] = _global_item_attribute(ClpRelation)

    dataset_groups: dict[str, DatasetGroupModel] = ib(
        factory=dict, converter=_load_dataset_groups, metadata=META
    )

    dataset: dict[str, DatasetModel]

    megacomplex: dict[str, Megacomplex] = ib(
        factory=dict,
        converter=lambda value: _load_model_items_from_dict(Megacomplex, value),
        metadata=META,
    )

    weights: list[Weight] = _global_item_attribute(Weight)

    @classmethod
    def create_class(cls, attributes: dict[str, Attribute]) -> type[Model]:
        """Create model class.

        Parameters
        ----------
        attributes: dict[str, Attribute]
            The model attributes.

        Returns
        -------
        type[Model]
        """
        cls_name = f"GlotaranModel_{str(uuid4()).replace('-','_')}"
        return make_class(cls_name, attributes, bases=(cls,))

    @classmethod
    def create_class_from_megacomplexes(
        cls, megacomplexes: Iterable[type[Megacomplex]]
    ) -> type[Model]:
        """Create model class for megacomplexes.

        Parameters
        ----------
        megacomplexes: list[type[Megacomplex]]
            The megacomplexes.

        Returns
        -------
        type[Model]
        """
        attributes: dict[str, Attribute] = {}
        dataset_types = set()
        for megacomplex in megacomplexes:
            if dataset_model_type := megacomplex.get_dataset_model_type():
                dataset_types |= {
                    dataset_model_type,
                }
            attributes |= _create_attributes_for_item(megacomplex)

        dataset_type = (
            DatasetModel
            if len(dataset_types) == 0
            else make_class(
                f"GlotaranDataset_{str(uuid4()).replace('-','_')}",
                [],
                bases=tuple(dataset_types),
                collect_by_mro=True,
            )
        )
        resolve_types(dataset_type)

        attributes.update(_create_attributes_for_item(dataset_type))

        attributes["dataset"] = _model_item_attribute(dataset_type)

        return cls.create_class(attributes)

    def as_dict(self) -> dict:
        """Get the model as dictionary.

        Returns
        -------
        dict
        """
        return asdict(
            self,
            recurse=True,
            retain_collection_types=True,
            filter=lambda attr, _: attr.name != "source_path",
        )

    def get_dataset_groups(self) -> dict[str, DatasetGroup]:
        """Get the dataset groups.

        Returns
        -------
        dict[str, DatasetGroup]

        Raises
        ------
        ModelError
            Raised if a dataset group is unknown.
        """
        groups = {}
        for dataset_model in self.dataset.values():
            group = dataset_model.group
            if group not in groups:
                try:
                    group_model = self.dataset_groups[group]
                except KeyError:
                    raise ModelError(f"Unknown dataset group '{group}'")
                groups[group] = DatasetGroup(  # type:ignore[call-arg]
                    residual_function=group_model.residual_function,
                    link_clp=group_model.link_clp,
                    model=self,
                )
            groups[group].dataset_models[dataset_model.label] = dataset_model
        return groups

    def iterate_items(self) -> Generator[tuple[str, dict[str, Item] | list[Item]], None, None]:
        """Iterate items.

        Yields
        ------
        tuple[str, dict[str, Item] | list[Item]]
            The name of the item and the individual items of the type.
        """
        for attr in fields(self.__class__):
            if META_ITEMS in attr.metadata:
                yield attr.name, getattr(self, attr.name)

    def iterate_all_items(self) -> Generator[Item, None, None]:
        """Iterate the individual items.

        Yields
        ------
        Item
            The individual item.
        """
        for _, items in self.iterate_items():
            yield from items.values() if isinstance(items, dict) else items

    def get_parameter_labels(self) -> set[str]:
        """Get all parameter labels.

        Returns
        -------
        set[str]
        """
        return {
            label
            for item in self.iterate_all_items()
            for _, label in iterate_parameter_names_and_labels(item)
        }

    def generate_parameters(self) -> Parameters:
        """Generate parameters for the model.

        Returns
        -------
        Parameters
            The generated parameters.

        .. # noqa: D414
        """
        return Parameters(
            {
                label: Parameter(label=label, value=0)  # type:ignore[call-arg]
                for label in self.get_parameter_labels()
            }
        )

    def get_issues(self, *, parameters: Parameters | None = None) -> list[ItemIssue]:
        """Get issues.

        Parameters
        ----------
        parameters: Parameters | None
            The parameters.

        Returns
        -------
        list[ItemIssue]
        """
        issues = []
        for item in self.iterate_all_items():
            issues += get_item_issues(item=item, model=self, parameters=parameters)
        return issues

    def validate(
        self, parameters: Parameters | None = None, raise_exception: bool = False
    ) -> MarkdownStr:
        """Get a string listing all issues in the model and missing parameters if specified.

        Parameters
        ----------
        parameters: Parameters | None
            The parameters.
        raise_exception: bool
            Whether to raise an exception on failed validation.

        Returns
        -------
        MarkdownStr

        Raises
        ------
        ModelError
            Raised if validation fails and raise_exception is true.
        """
        result = ""

        if issues := self.get_issues(parameters=parameters):
            result = f"Your model has {len(issues)} problem{'s' if len(issues) > 1 else ''}:\n"
            for issue in issues:
                result += f"\n * {issue.to_string()}"
            if raise_exception:
                raise ModelError(result)
        else:
            result = "Your model is valid."
        return MarkdownStr(result)

    def valid(self, parameters: Parameters | None = None) -> bool:
        """Check if the model is valid.

        Parameters
        ----------
        parameters: Parameters | None
            The parameters.

        Returns
        -------
        bool
        """
        return len(self.get_issues(parameters=parameters)) == 0

    def markdown(
        self,
        parameters: Parameters | None = None,
        initial_parameters: Parameters | None = None,
        base_heading_level: int = 1,
    ) -> MarkdownStr:
        """Format the model as Markdown string.

        Parameters will be included if specified.

        Parameters
        ----------
        parameters: Parameters | None
            Parameter to include.
        initial_parameters: Parameters | None
            Initial values for the parameters.
        base_heading_level: int
            Base heading level of the markdown sections.

            E.g.:

            - If it is 1 the string will start with '# Model'.
            - If it is 3 the string will start with '### Model'.

        Returns
        -------
        MarkdownStr
        """
        base_heading = "#" * base_heading_level
        string = f"{base_heading} Model\n\n"

        for name, items in self.iterate_items():
            if not items:
                continue

            string += f"{base_heading}# {name.replace('_', ' ').title()}\n\n"

            if isinstance(items, dict):
                items = items.values()  # type:ignore[assignment]
            for item in items:
                assert isinstance(item, Item)
                item_str = item_to_markdown(
                    item, parameters=parameters, initial_parameters=initial_parameters
                ).split("\n")
                string += f"- **{getattr(item, 'label', '&nbsp;')}**\n"
                for s in item_str[1:]:
                    string += f"{s}\n"
            string += "\n"
        return MarkdownStr(string)

    def _repr_markdown_(self) -> str:
        """Render ``ipython`` markdown.

        Returns
        -------
        str
        """
        return str(self.markdown(base_heading_level=3))



from __future__ import annotations

from pathlib import Path

import pytest
import xarray as xr

from glotaran.io import load_scheme
from glotaran.io import save_dataset
from glotaran.io import save_model
from glotaran.io import save_parameters
from glotaran.io import save_scheme
from glotaran.project import Scheme
from glotaran.testing.simulated_data.sequential_spectral_decay import DATASET
from glotaran.testing.simulated_data.sequential_spectral_decay import MODEL
from glotaran.testing.simulated_data.sequential_spectral_decay import PARAMETERS
from glotaran.utils.io import chdir_context

want = """\
model: m.yml
parameters: p.csv
data:
  dataset_1: d.nc
clp_link_tolerance: 0.0
clp_link_method: nearest
maximum_number_function_evaluations: null
add_svd: true
ftol: 1e-08
gtol: 1e-08
xtol: 1e-08
optimization_method: TrustRegionReflection
result_path: null
"""


@pytest.mark.parametrize("path_is_absolute", (True, False))
def test_save_scheme(tmp_path: Path, path_is_absolute: bool):
    save_model(MODEL, tmp_path / "m.yml")
    save_parameters(PARAMETERS, tmp_path / "p.csv")
    save_dataset(DATASET, tmp_path / "d.nc")
    scheme = Scheme(
        MODEL,
        PARAMETERS,
        {"dataset_1": DATASET},
    )
    if path_is_absolute is True:
        scheme_path = tmp_path / "testscheme.yml"
    else:
        scheme_path = Path("testscheme.yml")

    with chdir_context("." if path_is_absolute is True else tmp_path):
        save_scheme(file_name=scheme_path, format_name="yml", scheme=scheme)

        assert scheme_path.is_file()
        assert scheme_path.read_text() == want
        loaded = load_scheme(scheme_path)
        print(loaded.model.validate(loaded.parameters))
        assert loaded.model.valid(loaded.parameters)
        assert isinstance(scheme.data["dataset_1"], xr.Dataset)



"""This module contains the items."""
from __future__ import annotations

import contextlib
from inspect import getmro
from inspect import isclass
from textwrap import indent
from types import NoneType
from types import UnionType
from typing import TYPE_CHECKING
from typing import Any
from typing import Callable
from typing import ClassVar
from typing import Generator
from typing import Iterator
from typing import Type
from typing import TypeAlias
from typing import TypeVar
from typing import Union
from typing import get_args
from typing import get_origin

from attrs import NOTHING
from attrs import Attribute
from attrs import define
from attrs import evolve
from attrs import field
from attrs import fields
from attrs import resolve_types

from glotaran.parameter import Parameter
from glotaran.parameter import Parameters
from glotaran.utils.ipython import MarkdownStr

if TYPE_CHECKING:
    from glotaran.model.model import Model


META_ALIAS = "__glotaran_alias__"
META_VALIDATOR = "__glotaran_validator__"


class ItemIssue:
    """Baseclass for item issues."""

    def to_string(self) -> str:
        """Get the issue as string.

        Returns
        -------
        str

        .. # noqa: DAR202
        .. # noqa: DAR401
        """
        raise NotImplementedError

    def __rep__(self) -> str:
        """Get the representation."""
        return self.to_string()


class ModelItemIssue(ItemIssue):
    """Issue for missing model items."""

    def __init__(self, item_name: str, label: str):
        """Create a model issue.

        Parameters
        ----------
        item_name : str
            The name of the item.
        label : str
            The item label.
        """
        self._item_name = item_name
        self._label = label

    def to_string(self) -> str:
        """Get the issue as string.

        Returns
        -------
        str
        """
        return f"Missing model item '{self._item_name}' with label '{self._label}'."


class ParameterIssue(ItemIssue):
    """Issue for missing parameters."""

    def __init__(self, label: str):
        """Create a parameter issue.

        Parameters
        ----------
        label : str
            The parameter label.
        """
        self._label = label

    def to_string(self) -> str:
        """Get the issue as string.

        Returns
        -------
        str
        """
        return f"Missing parameter with label '{self._label}'."


@define(kw_only=True, slots=False)
class Item:
    """A baseclass for items."""


@define(kw_only=True, slots=False)
class ModelItem(Item):
    """An item with a label."""

    label: str


@define(kw_only=True, slots=False)
class TypedItem(Item):
    """An item with a type."""

    type: str
    __item_types__: ClassVar[dict[str, Type]]

    @classmethod
    def _register_item_class(cls):
        """Register a class as type."""
        item_type = cls.get_item_type()
        if item_type is not NOTHING:
            cls.__item_types__[item_type] = cls

    @classmethod
    def get_item_type(cls) -> str:
        """Get the type string.

        Returns
        -------
        str
        """
        return fields(cls).type.default

    @classmethod
    def get_item_types(cls) -> list[str]:
        """Get all type strings.

        Returns
        -------
        list[str]
        """
        return list(cls.__item_types__.keys())

    @classmethod
    def get_item_type_class(cls, item_type: str) -> Type:
        """Get the type for a type string.

        Parameters
        ----------
        item_type: str
            The type string.
        Returns
        -------
        Type
        """
        return cls.__item_types__[item_type]


@define(kw_only=True, slots=False)
class ModelItemTyped(TypedItem, ModelItem):
    """A model item with a type."""


ItemT = TypeVar("ItemT", bound="Item")
ModelItemT = TypeVar("ModelItemT", bound="ModelItem")

ParameterType: TypeAlias = Parameter | str
ModelItemType: TypeAlias = ModelItemT | str


def item_to_markdown(
    item: Item, parameters: Parameters | None = None, initial_parameters: Parameters | None = None
) -> MarkdownStr:
    """Get the item as markdown string.

    Parameters
    ----------
    item: Item
        The item.
    parameters: Parameters | None
        The parameters.
    initial_parameters: Parameters | None
        The initial parameters.

    Returns
    -------
    MarkdownStr
    """
    md = "\n"
    for attr in fields(item.__class__):
        name = attr.name
        value = getattr(item, name)
        if value is None:
            continue

        structure, item_type = strip_type_and_structure_from_attribute(attr)
        if item_type is Parameter and parameters is not None:
            if structure is dict:
                value = {
                    k: parameters.get(v.label if isinstance(v, Parameter) else v).markdown(
                        parameters, initial_parameters
                    )
                    for k, v in value.items()
                }
            elif structure is list:
                value = [
                    parameters.get(v.label if isinstance(v, Parameter) else v).markdown(
                        parameters, initial_parameters
                    )
                    for v in value
                ]
            else:
                value = parameters.get(
                    value.label if isinstance(value, Parameter) else value
                ).markdown(parameters, initial_parameters)

        property_md = indent(f"- _{name.replace('_', ' ').title()}_: {value}\n", "  ")

        md += property_md

    return MarkdownStr(md)


def iterate_attributes_of_type(
    item: type[Item], attr_type: type
) -> Generator[Attribute, None, None]:
    """Get attributes of type from an item type.

    Parameters
    ----------
    item: type[Item]
        The item type.
    attr_type: type
        The attribute type.

    Yields
    ------
    Attribute
        The attributes.
    """
    for attr in fields(item):
        _, item_type = strip_type_and_structure_from_attribute(attr)
        with contextlib.suppress(TypeError):
            # issubclass does for some reason not work with e.g. tuple as item_type
            # and Parameter as attr_type
            if isclass(item_type) and issubclass(item_type, attr_type):
                yield attr


def model_attributes(
    item: type[Item], with_alias: bool = True
) -> Generator[Attribute, None, None]:
    """Get model attributes from an item type.

    Parameters
    ----------
    item: type[Item]
        The item type.
    with_alias: bool
        Whether to return aliased attributes.

    Yields
    ------
    Attribute
        The model attributes.
    """
    for attr in iterate_attributes_of_type(item, ModelItem):
        if with_alias or META_ALIAS not in attr.metadata:
            yield attr


def parameter_attributes(item: type[Item]) -> Generator[Attribute, None, None]:
    """Get parameter attributes from an item type.

    Parameters
    ----------
    item: type[Item]
        The item type.

    Yields
    ------
    Attribute
        The parameter attributes.
    """
    yield from iterate_attributes_of_type(item, Parameter)


def iterate_names_and_labels(
    item: Item, attributes: Generator[Attribute, None, None]
) -> Generator[tuple[str, str], None, None]:
    """Get attribute names and labels.

    Parameters
    ----------
    item: Item
        The item.
    attributes: Generator[Attribute, None, None]
        The attributes.

    Yields
    ------
    tuple[str, str]
        The name and the label.
    """
    for attr in attributes:
        structure, _ = strip_type_and_structure_from_attribute(attr)
        value = getattr(item, attr.name)
        name: str = attr.metadata.get(META_ALIAS, attr.name)

        if not value:
            continue

        if structure is dict:
            for v in value.values():
                yield name, v if isinstance(v, str) else (name, v.label)  # type:ignore[misc]

        elif structure is list:
            for v in value:
                yield name, v if isinstance(v, str) else (name, v.label)  # type:ignore[misc]

        else:
            yield name, value if isinstance(value, str) else (
                name,
                value.label,  # type:ignore[misc]
            )


def iterate_model_item_names_and_labels(item: Item) -> Generator[tuple[str, str], None, None]:
    """Get model item names and labels.

    Parameters
    ----------
    item: Item
        The item.

    Yields
    ------
    tuple[str, str]
        The name and the label.
    """
    yield from iterate_names_and_labels(item, model_attributes(item.__class__))


def iterate_parameter_names_and_labels(item: Item) -> Generator[tuple[str, str], None, None]:
    """Get parameter item names and labels.

    Parameters
    ----------
    item: Item
        The item.

    Yields
    ------
    tuple[str, str]
        The name and the label.
    """
    yield from iterate_names_and_labels(item, parameter_attributes(item.__class__))


def fill_item_attributes(
    item: Item,
    iterator: Iterator[Attribute],
    fill_function: Callable[[str, str], Parameter | ModelItem],
):
    """Fill item attributes.

    Parameters
    ----------
    item: Item
        The item.
    iterator: Iterator[Attribute]
        An iterator over attributes.
    fill_function: Callable[[str, str], Parameter | ModelItem]
        The function to fill the values.
    """
    for attr in iterator:
        value = getattr(item, attr.name)
        if not value:
            continue

        structure, _ = strip_type_and_structure_from_attribute(attr)
        name = attr.metadata.get(META_ALIAS, attr.name)
        if structure is dict:
            value = {
                k: fill_function(name, v) if isinstance(v, str) else fill_function(name, v.label)
                for k, v in value.items()
            }
        elif structure is list:
            value = [
                fill_function(name, v) if isinstance(v, str) else fill_function(name, v.label)
                for v in value
            ]
        else:
            value = (
                fill_function(name, value)
                if isinstance(value, str)
                else fill_function(name, value.label)
            )

        setattr(item, attr.name, value)


def fill_item(item: ItemT, model: Model, parameters: Parameters) -> ItemT:
    """Fill an item.

    Parameters
    ----------
    item: ItemT
        The item.
    model: Model
        The model.
    parameters: Parameters
        The parameters.

    Returns
    -------
    ItemT
        The filled item.
    """
    item = evolve(item)
    fill_item_model_attributes(item, model, parameters)
    fill_item_parameter_attributes(item, parameters)
    return item


def fill_item_model_attributes(item: Item, model: Model, parameters: Parameters):
    """Fill item model attributes.

    Parameters
    ----------
    item: Item
        The item.
    model: Model
        The model.
    parameters: Parameters
        The parameters.
    """
    fill_item_attributes(
        item,
        model_attributes(item.__class__),
        lambda name, label: fill_item(getattr(model, name)[label], model, parameters),
    )


def fill_item_parameter_attributes(item: Item, parameters: Parameters):
    """Fill item parameter attributes.

    Parameters
    ----------
    item: Item
        The item.
    parameters: Parameters
        The parameters.
    """
    fill_item_attributes(
        item, parameter_attributes(item.__class__), lambda _, label: parameters.get(label)
    )


def get_item_model_issues(item: Item, model: Model) -> list[ItemIssue]:
    """Get model item issues for an item.

    Parameters
    ----------
    item: Item
        The item.
    model: Model
        The model.

    Returns
    -------
    list[ItemIssue]
    """
    return [
        ModelItemIssue(name, label)
        for name, label in iterate_model_item_names_and_labels(item)
        if label not in getattr(model, name)
    ]


def get_item_parameter_issues(item: Item, parameters: Parameters) -> list[ItemIssue]:
    """Get model item issues for an item.

    Parameters
    ----------
    item: Item
        The item.
    parameters: Parameters
        The parameters.

    Returns
    -------
    list[ItemIssue]
    """
    return [
        ParameterIssue(label)
        for name, label in iterate_parameter_names_and_labels(item)
        if not parameters.has(label)
    ]


def get_item_validator_issues(
    item: Item, model: Model, parameters: Parameters | None = None
) -> list[ItemIssue]:
    """Get validator issues for an item.

    Parameters
    ----------
    item: Item
        The item.
    model: Model
        The model.
    parameters: Parameters | None
        The parameters.

    Returns
    -------
    list[ItemIssue]
    """
    issues = []
    for name, validator in [
        (attr.name, attr.metadata[META_VALIDATOR])
        for attr in fields(item.__class__)
        if META_VALIDATOR in attr.metadata
    ]:
        issues += validator(getattr(item, name), item, model, parameters)

    return issues


def get_item_issues(
    *, item: Item, model: Model, parameters: Parameters | None = None
) -> list[ItemIssue]:
    """Get issues for an item.

    Parameters
    ----------
    item: Item
        The item.
    model: Model
        The model.
    parameters: Parameters | None
        The parameters.

    Returns
    -------
    list[ItemIssue]
    """
    issues = get_item_model_issues(item, model)
    issues += get_item_validator_issues(item, model, parameters)
    if parameters is not None:
        issues += get_item_parameter_issues(item, parameters)
    return issues


def strip_type_and_structure_from_attribute(attr: Attribute) -> tuple[None | list | dict, type]:
    """Strip the type and the structure from an attribute.

    Parameters
    ----------
    attr: Attribute
        The attribute.

    Returns
    -------
    tuple[None | list | dict, type]:
        The structure and the type.
    """
    definition = attr.type
    definition = strip_option_type(definition)
    structure, definition = strip_structure_type(definition)
    definition = strip_option_type(definition, strip_type=str)
    return structure, definition


def strip_option_type(definition: type, strip_type: type = NoneType) -> type:
    """Strip the type if the definition is an option.

    Parameters
    ----------
    definition: type
        The definition.
    strip_type: type
        The type which should be removed from the option.

    Returns
    -------
    type
    """
    args = list(get_args(definition))
    if get_origin(definition) in [Union, UnionType] and strip_type in args:
        args.remove(strip_type)
        definition = args[0]
    return definition


def strip_structure_type(definition: type) -> tuple[None | list | dict, type]:
    """Strip the structure from a definition.

    Parameters
    ----------
    definition: type
        The definition.

    Returns
    -------
    tuple[None | list | dict, type]:
        The structure and the type.
    """
    structure = get_origin(definition)
    if structure is list:
        definition = get_args(definition)[0]
    elif structure is dict:
        definition = get_args(definition)[1]
    else:
        structure = None

    return structure, definition


def item(cls: type[ItemT]) -> type[ItemT]:
    """Create an item from a class.

    Parameters
    ----------
    cls: type[ItemT]
        The class.

    Returns
    -------
    type[ItemT]
    """
    parent = getmro(cls)[1]
    cls = define(kw_only=True, slots=False)(cls)
    if parent in (TypedItem, ModelItemTyped):
        assert issubclass(cls, TypedItem)
        cls.__item_types__ = {}
    elif issubclass(cls, TypedItem):
        cls._register_item_class()
    resolve_types(cls)
    return cls


def attribute(
    *,
    alias: str | None = None,
    default: Any = NOTHING,
    factory: Callable[[], Any] | None = None,
    validator: Callable[[Any, Item, Model, Parameters | None], list[ItemIssue]] | None = None,
) -> Attribute:
    """Create an attribute for an item.

    Parameters
    ----------
    alias: str | None
        The alias of the attribute (only useful for model items).
    default: Any
        The default value of the attribute.
    factory: Callable[[], Any] | None
        A factory function for the attribute.
    validator: Callable[[Any, Item, Model, Parameters | None], list[ItemIssue]] | None
        A validator function for the attribute.

    Returns
    -------
    Attribute
    """
    metadata: dict[str, Any] = {}
    if alias is not None:
        metadata[META_ALIAS] = alias
    if validator is not None:
        metadata[META_VALIDATOR] = validator
    return field(default=default, factory=factory, metadata=metadata)



"""This module contains clp constraint items."""
from __future__ import annotations

from glotaran.model.interval_item import IntervalItem
from glotaran.model.item import TypedItem
from glotaran.model.item import item


@item
class ClpConstraint(TypedItem, IntervalItem):
    """Baseclass for clp constraints.

    There are two types: zero and equal. See the documentation of
    the respective classes for details.
    """

    target: str


@item
class ZeroConstraint(ClpConstraint):
    """Constraints the target to 0 in the given interval."""

    type: str = "zero"


@item
class OnlyConstraint(ZeroConstraint):
    """Constraints the target to 0 outside the given interval."""

    type: str = "only"

    def applies(self, index: float | None) -> bool:
        """Check if the constraint applies on this index.

        Parameters
        ----------
        index : float
            The index.

        Returns
        -------
        bool
        """
        return not super().applies(index)



"""This package contains builtin plugins."""
from glotaran.deprecation.deprecation_utils import deprecate_submodule

read_data_file = deprecate_submodule(
    deprecated_module_name="glotaran.builtin.read_data_file",
    new_module_name="glotaran.builtin.io",
    to_be_removed_in_version="0.6.0",
)



from __future__ import annotations

import numpy as np

from glotaran.optimization.test.models import DecayModel
from glotaran.parameter import Parameters


class OneCompartmentDecay:
    scale = 2
    wanted_parameters = Parameters.from_list([101e-4])
    initial_parameters = Parameters.from_list([100e-5, [scale, {"vary": False}]])

    global_axis = np.asarray([1.0])
    model_axis = np.arange(0, 150, 1.5)

    sim_model_dict = {
        "megacomplex": {
            "m1": {"type": "simple-kinetic-test-mc", "is_index_dependent": False},
            "m2": {"type": "simple-spectral-test-mc"},
        },
        "dataset": {
            "dataset1": {
                "megacomplex": ["m1"],
                "global_megacomplex": ["m2"],
                "kinetic": ["1"],
            }
        },
    }
    sim_model = DecayModel(**sim_model_dict)
    model_dict = {
        "megacomplex": {"m1": {"type": "simple-kinetic-test-mc", "is_index_dependent": False}},
        "dataset": {
            "dataset1": {
                "megacomplex": ["m1"],
                "kinetic": ["1"],
                "scale": "2",
            }
        },
    }
    model_dict["dataset"]["dataset1"]["scale"] = "2"  # type:ignore[index]
    model = DecayModel(**model_dict)


class TwoCompartmentDecay:
    wanted_parameters = Parameters.from_list([11e-4, 22e-5])
    initial_parameters = Parameters.from_list([10e-4, 20e-5])

    global_axis = np.asarray([1.0, 2])
    model_axis = np.arange(0, 150, 1.5)

    sim_model = DecayModel(
        **{
            "megacomplex": {
                "m1": {"type": "simple-kinetic-test-mc", "is_index_dependent": False},
                "m2": {"type": "simple-spectral-test-mc"},
            },
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                    "global_megacomplex": ["m2"],
                    "kinetic": ["1", "2"],
                }
            },
        }
    )
    model = DecayModel(
        **{
            "megacomplex": {"m1": {"type": "simple-kinetic-test-mc", "is_index_dependent": False}},
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                    "kinetic": ["1", "2"],
                }
            },
        }
    )


class ThreeDatasetDecay:
    wanted_parameters = Parameters.from_list([101e-4, 201e-3])
    initial_parameters = Parameters.from_list([100e-5, 200e-3])

    global_axis = np.asarray([1.0])
    model_axis = np.arange(0, 150, 1.5)

    global_axis2 = np.asarray([1.0, 2.01])
    model_axis2 = np.arange(0, 100, 1.5)

    global_axis3 = np.asarray([0.99, 3.0])
    model_axis3 = np.arange(0, 150, 1.5)

    sim_model_dict = {
        "megacomplex": {
            "m1": {"type": "simple-kinetic-test-mc", "is_index_dependent": False},
            "m2": {"type": "simple-spectral-test-mc"},
        },
        "dataset": {
            "dataset1": {
                "megacomplex": ["m1"],
                "global_megacomplex": ["m2"],
                "kinetic": ["1"],
            },
            "dataset2": {
                "megacomplex": ["m1"],
                "global_megacomplex": ["m2"],
                "kinetic": ["1", "2"],
            },
            "dataset3": {
                "megacomplex": ["m1"],
                "global_megacomplex": ["m2"],
                "kinetic": ["2"],
            },
        },
    }
    sim_model = DecayModel(**sim_model_dict)

    model_dict = {
        "megacomplex": {"m1": {"type": "simple-kinetic-test-mc", "is_index_dependent": False}},
        "dataset": {
            "dataset1": {"megacomplex": ["m1"], "kinetic": ["1"]},
            "dataset2": {
                "megacomplex": ["m1"],
                "kinetic": ["1", "2"],
            },
            "dataset3": {"megacomplex": ["m1"], "kinetic": ["2"]},
        },
    }
    model = DecayModel(**model_dict)


class MultichannelMulticomponentDecay:
    wanted_parameters = Parameters.from_dict(
        {
            "k": [0.006, 0.003, 0.0003, 0.03],
            "loc": [
                ["1", 14705],
                ["2", 13513],
                ["3", 14492],
                ["4", 14388],
            ],
            "amp": [
                ["1", 1],
                ["2", 2],
                ["3", 5],
                ["4", 20],
            ],
            "del": [
                ["1", 400],
                ["2", 100],
                ["3", 300],
                ["4", 200],
            ],
        }
    )
    initial_parameters = Parameters.from_dict({"k": [0.006, 0.003, 0.0003, 0.03]})

    global_axis = np.arange(12820, 15120, 50)
    model_axis = np.arange(0, 150, 1.5)

    sim_model = DecayModel(
        **{
            "megacomplex": {
                "m1": {"type": "simple-kinetic-test-mc", "is_index_dependent": False},
                "m2": {
                    "type": "shaped-spectral-test-mc",
                    "location": ["loc.1", "loc.2", "loc.3", "loc.4"],
                    "delta": ["del.1", "del.2", "del.3", "del.4"],
                    "amplitude": ["amp.1", "amp.2", "amp.3", "amp.4"],
                },
            },
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                    "global_megacomplex": ["m2"],
                    "kinetic": ["k.1", "k.2", "k.3", "k.4"],
                }
            },
        }
    )
    model = DecayModel(
        **{
            "megacomplex": {"m1": {"type": "simple-kinetic-test-mc", "is_index_dependent": False}},
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                    "kinetic": ["k.1", "k.2", "k.3", "k.4"],
                }
            },
        }
    )


class FullModel:
    model = DecayModel(
        **{
            "megacomplex": {
                "m1": {"type": "simple-kinetic-test-mc", "is_index_dependent": False},
                "m2": {
                    "type": "shaped-spectral-test-mc",
                    "location": ["loc.1", "loc.2", "loc.3", "loc.4"],
                    "delta": ["del.1", "del.2", "del.3", "del.4"],
                    "amplitude": ["amp.1", "amp.2", "amp.3", "amp.4"],
                },
            },
            "dataset": {
                "dataset1": {
                    "megacomplex": ["m1"],
                    "global_megacomplex": ["m2"],
                    "kinetic": ["k.1", "k.2", "k.3", "k.4"],
                }
            },
        }
    )
    parameters = Parameters.from_dict(
        {
            "k": [0.006, 0.003, 0.0003, 0.03],
            "loc": [
                ["1", 14705],
                ["2", 13513],
                ["3", 14492],
                ["4", 14388],
            ],
            "amp": [
                ["1", 1],
                ["2", 2],
                ["3", 5],
                ["4", 20],
            ],
            "del": [
                ["1", 400],
                ["2", 100],
                ["3", 300],
                ["4", 200],
            ],
        }
    )
    global_axis = np.arange(12820, 15120, 50)
    model_axis = np.arange(0, 150, 1.5)
    coordinates = {"global": global_axis, "model": model_axis}



from __future__ import annotations

from pathlib import Path
from textwrap import dedent
from typing import TYPE_CHECKING

import pytest

from glotaran.builtin.io.pandas.csv import CsvProjectIo
from glotaran.builtin.io.yml.yml import YmlProjectIo
from glotaran.io import ProjectIoInterface
from glotaran.parameter import Parameters
from glotaran.plugin_system.base_registry import PluginOverwriteWarning
from glotaran.plugin_system.base_registry import __PluginRegistry
from glotaran.plugin_system.project_io_registration import SAVING_OPTIONS_DEFAULT
from glotaran.plugin_system.project_io_registration import SavingOptions
from glotaran.plugin_system.project_io_registration import get_project_io
from glotaran.plugin_system.project_io_registration import get_project_io_method
from glotaran.plugin_system.project_io_registration import is_known_project_format
from glotaran.plugin_system.project_io_registration import known_project_formats
from glotaran.plugin_system.project_io_registration import load_model
from glotaran.plugin_system.project_io_registration import load_parameters
from glotaran.plugin_system.project_io_registration import load_result
from glotaran.plugin_system.project_io_registration import load_scheme
from glotaran.plugin_system.project_io_registration import project_io_plugin_table
from glotaran.plugin_system.project_io_registration import register_project_io
from glotaran.plugin_system.project_io_registration import save_model
from glotaran.plugin_system.project_io_registration import save_parameters
from glotaran.plugin_system.project_io_registration import save_result
from glotaran.plugin_system.project_io_registration import save_scheme
from glotaran.plugin_system.project_io_registration import set_project_plugin
from glotaran.plugin_system.project_io_registration import show_project_io_method_help
from glotaran.plugin_system.project_io_registration import supported_file_extensions_project_io
from glotaran.testing.plugin_system import monkeypatch_plugin_registry_project_io

if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Sequence
    from typing import Any

    from _pytest.capture import CaptureFixture

    from glotaran.model import Model
    from glotaran.project import Result
    from glotaran.project import Scheme
    from glotaran.typing import StrOrPath


class MockFileLoadable:
    source_path = "bar"
    func_args: dict[str, Any] = {}


class MockProjectIo(ProjectIoInterface):
    # TODO: Investigate why write methods raises an [override] type error and load functions don't
    def load_model(self, file_name: StrOrPath, **kwargs: Any) -> Model:
        """This docstring is just for help testing of 'load_model'."""
        mock_obj = MockFileLoadable()
        mock_obj.func_args = {"file_name": file_name, **kwargs}
        return mock_obj  # type:ignore[return-value]

    def save_model(self, model: Model, file_name: StrOrPath, **kwargs: Any):
        model.func_args.update(  # type:ignore[attr-defined]
            **{
                "file_name": file_name,
                "data_object": model,
                **kwargs,
            }
        )

    def load_parameters(self, file_name: StrOrPath, **kwargs: Any) -> Parameters:
        mock_obj = MockFileLoadable()
        mock_obj.func_args = {"file_name": file_name, **kwargs}
        return mock_obj  # type:ignore[return-value]

    def save_parameters(self, parameters: Parameters, file_name: StrOrPath, **kwargs: Any):
        parameters.func_args.update(  # type:ignore[attr-defined]
            **{
                "file_name": file_name,
                "data_object": parameters,
                **kwargs,
            }
        )

    def load_scheme(self, file_name: StrOrPath, **kwargs: Any) -> Scheme:
        mock_obj = MockFileLoadable()
        mock_obj.func_args = {"file_name": file_name, **kwargs}
        return mock_obj  # type:ignore[return-value]

    def save_scheme(self, scheme: Scheme, file_name: StrOrPath, **kwargs: Any):
        scheme.func_args.update(  # type:ignore[attr-defined]
            **{
                "file_name": file_name,
                "data_object": scheme,
                **kwargs,
            }
        )

    def load_result(self, result_path: StrOrPath, **kwargs: Any) -> Result:
        mock_obj = MockFileLoadable()
        mock_obj.func_args = {"file_name": result_path, **kwargs}
        return mock_obj  # type:ignore[return-value]

    def save_result(
        self,
        result: Result,
        result_path: StrOrPath,
        *,
        saving_options: SavingOptions = SAVING_OPTIONS_DEFAULT,
        **kwargs: Any,
    ):
        result.func_args.update(  # type:ignore[attr-defined]
            **{
                "file_name": result_path,
                "data_object": result,
                **kwargs,
            }
        )


class MockProjectIoPartial(ProjectIoInterface):
    def load_model(self, file_name: StrOrPath, **kwargs: Any) -> Model:
        pass

    def load_parameters(self, file_name: StrOrPath, **kwargs: Any) -> Parameters:
        pass

    def load_scheme(self, file_name: StrOrPath, **kwargs: Any) -> Scheme:
        pass

    def load_result(self, result_path: StrOrPath, **kwargs: Any) -> Result:
        pass


MOCK_REGISTRY_VALUES = {
    "foo": ProjectIoInterface("foo"),
    "mock": MockProjectIo("bar"),
    "test_project_io_registration.MockProjectIo_bar": MockProjectIo("bar"),
}


@pytest.fixture
def mocked_registry():
    with monkeypatch_plugin_registry_project_io(MOCK_REGISTRY_VALUES, create_new_registry=True):
        yield


@pytest.mark.usefixtures("mocked_registry")
def test_register_project_io():
    """Registered project_io plugin is in registry"""

    @register_project_io("dummy")
    class Dummy(ProjectIoInterface):
        pass

    @register_project_io(["dummy2", "dummy3"])
    class Dummy2(ProjectIoInterface):
        pass

    for format_name, plugin_class in [("dummy", Dummy), ("dummy2", Dummy2), ("dummy3", Dummy2)]:
        assert format_name in __PluginRegistry.project_io
        assert isinstance(__PluginRegistry.project_io[format_name], plugin_class)
        assert isinstance(
            __PluginRegistry.project_io[
                f"test_project_io_registration.{plugin_class.__name__}_{format_name}"
            ],
            plugin_class,
        )
        assert __PluginRegistry.project_io[format_name].format == format_name


@pytest.mark.usefixtures("mocked_registry")
def test_register_project_io_warning():
    """PluginOverwriteWarning raised pointing to correct file."""

    with pytest.warns(PluginOverwriteWarning, match="Dummy.+dummy.+Dummy2") as record:

        @register_project_io("dummy")
        class Dummy(ProjectIoInterface):
            pass

        @register_project_io("dummy")
        class Dummy2(ProjectIoInterface):
            pass

        assert len(record) == 1
        assert Path(record[0].filename) == Path(__file__)


@pytest.mark.usefixtures("mocked_registry")
def test_known_project_format():
    """Known format in mocked register"""
    assert is_known_project_format("foo")
    assert is_known_project_format("mock")
    assert not is_known_project_format("baz")


def test_known_project_format_actual_register():
    """Builtins in are in actual register."""
    assert is_known_project_format("yml")
    assert is_known_project_format("yaml")
    assert is_known_project_format("yml_str")
    assert is_known_project_format("csv")
    assert is_known_project_format("glotaran.builtin.io.yml.yml.YmlProjectIo_yml")
    assert is_known_project_format("glotaran.builtin.io.yml.yml.YmlProjectIo_yaml")
    assert is_known_project_format("glotaran.builtin.io.yml.yml.YmlProjectIo_yml_str")
    assert is_known_project_format("glotaran.builtin.io.pandas.csv.CsvProjectIo_csv")


@pytest.mark.parametrize(
    "format_name, io_class",
    (
        ("yml", YmlProjectIo),
        ("yaml", YmlProjectIo),
        ("yml_str", YmlProjectIo),
        ("csv", CsvProjectIo),
    ),
)
def test_get_project_io(format_name: str, io_class: type[ProjectIoInterface]):
    """Get the right instance"""
    assert isinstance(get_project_io(format_name), io_class)
    assert get_project_io(format_name).format == format_name


@pytest.mark.usefixtures("mocked_registry")
def test_known_project_formats():
    """Known formats are the same as mocked register keys"""
    assert known_project_formats() == ["foo", "mock"]


@pytest.mark.usefixtures("mocked_registry")
def test_set_project_plugin():
    """Set Change Plugin used for format foo"""
    assert isinstance(get_project_io("foo"), ProjectIoInterface)
    set_project_plugin("foo", "test_project_io_registration.MockProjectIo_bar")
    assert isinstance(get_project_io("foo"), MockProjectIo)


@pytest.mark.parametrize(
    "load_function",
    (
        load_model,
        load_parameters,
        load_scheme,
        load_result,
    ),
)
@pytest.mark.usefixtures("mocked_registry")
def test_load_functions(tmp_path: Path, load_function: Callable[..., Any]):
    """All args and kwargs are passes correctly."""
    file_path = tmp_path / "model.mock"
    file_path.touch()

    result = load_function(file_path, dummy_arg="baz")

    assert result.func_args == {"file_name": file_path.as_posix(), "dummy_arg": "baz"}
    assert result.source_path == Path(file_path).as_posix()


@pytest.mark.parametrize("sub_dir", ("", "sub_dir"))
@pytest.mark.parametrize(
    "save_function",
    (
        save_model,
        save_parameters,
        save_scheme,
        save_result,
    ),
)
@pytest.mark.parametrize("update_source_path", (True, False))
@pytest.mark.usefixtures("mocked_registry")
def test_write_functions(
    tmp_path: Path, save_function: Callable[..., Any], update_source_path: bool, sub_dir: str
):
    """All args and kwargs are passes correctly."""
    file_path = tmp_path / "sub_dir" / "model.mock"
    mock_obj = MockFileLoadable()

    save_function(
        mock_obj, file_path, "mock", update_source_path=update_source_path, dummy_arg="baz"
    )

    assert mock_obj.func_args == {
        "file_name": file_path.as_posix(),
        "data_object": mock_obj,
        "dummy_arg": "baz",
    }
    if update_source_path is True:
        assert mock_obj.source_path == file_path.as_posix()
    else:
        assert mock_obj.source_path == "bar"


@pytest.mark.parametrize(
    "load_function, error_regex",
    (
        (load_model, "read models"),
        (load_parameters, "read parameters"),
        (load_scheme, "read scheme"),
        (load_result, "read result"),
    ),
)
@pytest.mark.usefixtures("mocked_registry")
def test_load_functions_value_error(
    tmp_path: Path, load_function: Callable[..., Any], error_regex: str
):
    """Raise ValueError if load method isn't implemented."""
    file_path = tmp_path / "dummy.foo"

    with pytest.raises(ValueError, match=f"Cannot {error_regex} with format 'foo'"):
        load_function(file_path, "foo")


@pytest.mark.parametrize(
    "save_function, error_regex",
    (
        (save_model, "save models"),
        (save_parameters, "save parameters"),
        (save_scheme, "save scheme"),
        (save_result, "save result"),
    ),
)
@pytest.mark.usefixtures("mocked_registry")
def test_save_functions_value_error(
    tmp_path: Path, save_function: Callable[..., Any], error_regex: str
):
    """Raise ValueError if save method isn't implemented."""
    file_path = tmp_path / "dummy.foo"
    mock_obj = MockFileLoadable()

    with pytest.raises(ValueError, match=f"Cannot {error_regex} with format 'foo'"):
        save_function(mock_obj, file_path)


@pytest.mark.parametrize(
    "function",
    (save_model, save_parameters, save_scheme, save_result),
)
@pytest.mark.usefixtures("mocked_registry")
def test_protect_from_overwrite_save_functions(tmp_path: Path, function: Callable[..., Any]):
    """Raise FileExistsError if file exists."""

    file_path = tmp_path / "dummy.foo"
    file_path.touch()

    with pytest.raises(FileExistsError, match="The file .+? already exists"):
        function("foo", file_path, "bar")


@pytest.mark.usefixtures("mocked_registry")
def test_get_project_io_method():
    """Methods have the same code."""
    io = get_project_io("mock")
    result = get_project_io_method("mock", "load_model")

    assert result.__code__ == io.load_model.__code__


@pytest.mark.usefixtures("mocked_registry")
def test_show_project_io_method_help(capsys: CaptureFixture):
    """Same help as when called directly."""
    plugin = MockProjectIo("foo")
    help(plugin.load_model)
    original_help, _ = capsys.readouterr()

    show_project_io_method_help(format_name="mock", method_name="load_model")
    result, _ = capsys.readouterr()

    assert "This docstring is just for help testing of 'load_model'." in result
    assert result == original_help


@pytest.mark.usefixtures("mocked_registry")
def test_project_io_plugin_table():
    """Plugin foo supports no function and mock supports all"""
    expected = dedent(
        """\
        |  __Format name__  |  __load_model__  |  __save_model__  |  __load_parameters__  |  __save_parameters__  |  __load_scheme__  |  __save_scheme__  |  __load_result__  |  __save_result__  |
        |-------------------|------------------|------------------|-----------------------|-----------------------|-------------------|-------------------|-------------------|-------------------|
        |       `foo`       |        /         |        /         |           /           |           /           |         /         |         /         |         /         |         /         |
        |      `mock`       |        *         |        *         |           *           |           *           |         *         |         *         |         *         |         *         |
        """  # noqa: E501
    )

    assert f"{project_io_plugin_table()}\n" == expected


@pytest.mark.usefixtures("mocked_registry")
def test_project_io_plugin_table_full():
    """Full Table with all extras"""
    expected = dedent(
        """\
        |                 __Format name__                  |  __load_model__  |  __save_model__  |  __load_parameters__  |  __save_parameters__  |  __load_scheme__  |  __save_scheme__  |  __load_result__  |  __save_result__  |                  __Plugin name__                  |
        |--------------------------------------------------|------------------|------------------|-----------------------|-----------------------|-------------------|-------------------|-------------------|-------------------|---------------------------------------------------|
        |                      `foo`                       |        /         |        /         |           /           |           /           |         /         |         /         |         /         |         /         |  `glotaran.io.interface.ProjectIoInterface_foo`   |
        |                      `mock`                      |        *         |        *         |           *           |           *           |         *         |         *         |         *         |         *         | `test_project_io_registration.MockProjectIo_mock` |
        | `test_project_io_registration.MockProjectIo_bar` |        *         |        *         |           *           |           *           |         *         |         *         |         *         |         *         | `test_project_io_registration.MockProjectIo_bar`  |
        """  # noqa: E501
    )

    assert f"{project_io_plugin_table(plugin_names=True,full_names=True)}\n" == expected


@pytest.mark.parametrize(
    "method_names, expected",
    (
        (
            "load_model",
            [".mock", ".mock_partial"],
        ),
        (
            "save_model",
            [".mock"],
        ),
        (
            ["load_model", "save_model"],
            [".mock"],
        ),
    ),
)
def test_supported_file_extensions_project_io(
    method_names: str | Sequence[str], expected: list[str]
):
    """Extension don't list full plugin name and omit extension that don't support all methods."""

    with monkeypatch_plugin_registry_project_io(
        {**MOCK_REGISTRY_VALUES, "mock_partial": MockProjectIoPartial}, create_new_registry=True
    ):
        assert list(supported_file_extensions_project_io(method_names)) == expected



from pathlib import Path

import pytest

from glotaran.model.dataset_model import DatasetModel
from glotaran.model.item import ModelItem
from glotaran.model.item import ModelItemType
from glotaran.model.item import ParameterType
from glotaran.model.item import item
from glotaran.model.megacomplex import Megacomplex
from glotaran.model.megacomplex import megacomplex
from glotaran.model.model import DEFAULT_DATASET_GROUP
from glotaran.model.model import Model


@item
class MockItemSimple(ModelItem):
    param: ParameterType
    param_list: list[ParameterType]
    param_dict: dict[tuple[str, str], ParameterType]
    megacomplex: ModelItemType[Megacomplex]
    number: int = 42


@megacomplex()
class MockMegacomplexSimple(Megacomplex):
    type: str = "simple"
    dimension: str = "model"
    test_item: ModelItemType[MockItemSimple] | None


@megacomplex()
class MockMegacomplexItemList(Megacomplex):
    type: str = "list"
    dimension: str = "model"
    test_item_in_list: list[ModelItemType[MockItemSimple]]


@megacomplex()
class MockMegacomplexItemDict(Megacomplex):
    type: str = "dict"
    dimension: str = "model"
    test_item_in_dict: dict[str, ModelItemType[MockItemSimple]]


@item
class MockDatasetModel(DatasetModel):
    test_item_dataset: ModelItemType[MockItemSimple]
    test_property_dataset1: int
    test_property_dataset2: ParameterType


@megacomplex(dataset_model_type=MockDatasetModel)
class MockMegacomplexWithDataset(Megacomplex):
    type: str = "dataset"
    dimension: str = "model"


@megacomplex(unique=True)
class MockMegacomplexUnique(Megacomplex):
    type: str = "unique"
    dimension: str = "model"


@megacomplex(exclusive=True)
class MockMegacomplexExclusive(Megacomplex):
    type: str = "exclusive"
    dimension: str = "model"


@pytest.fixture
def test_model_dict():
    model_dict = {
        "megacomplex": {
            "m1": {"type": "simple", "test_item": "t2"},
            "m2": {"type": "dataset", "dimension": "model2"},
        },
        "dataset_groups": {
            "testgroup": {"residual_function": "non_negative_least_squares", "link_clp": True}
        },
        "weights": [
            {
                "datasets": ["d1", "d2"],
                "global_interval": (1, 4),
                "model_interval": (2, 3),
                "value": 5.4,
            }
        ],
        "test_item": {
            "t1": {
                "param": "foo",
                "megacomplex": "m1",
                "param_list": ["bar", "baz"],
                "param_dict": {("s1", "s2"): "baz"},
            },
            "t2": {
                "param": "baz",
                "megacomplex": "m2",
                "param_list": ["foo"],
                "param_dict": {},
                "number": 7,
            },
        },
        "dataset": {
            "dataset1": {
                "megacomplex": ["m1"],
                "scale": "scale_1",
                "test_item_dataset": "t1",
                "test_property_dataset1": 1,
                "test_property_dataset2": "bar",
            },
            "dataset2": {
                "megacomplex": ["m2"],
                "global_megacomplex": ["m1"],
                "scale": "scale_2",
                "test_item_dataset": "t2",
                "test_property_dataset1": 1,
                "test_property_dataset2": "bar",
                "group": "testgroup",
            },
        },
    }
    model_dict["test_item_dataset"] = model_dict["test_item"]
    return model_dict


@pytest.fixture
def test_model(test_model_dict):
    mcls = Model.create_class_from_megacomplexes(
        [MockMegacomplexSimple, MockMegacomplexWithDataset]
    )
    return mcls(**test_model_dict)


def test_model_create_class():
    m = Model.create_class([])(dataset={})
    print(m)
    assert DEFAULT_DATASET_GROUP in m.dataset_groups

    m = Model.create_class([])(
        **{
            "dataset": {},
            "dataset_groups": {
                "test": {"residual_function": "non_negative_least_squares", "link_clp": False}
            },
        }
    )
    print(m)
    assert DEFAULT_DATASET_GROUP in m.dataset_groups
    assert "test" in m.dataset_groups
    assert m.dataset_groups["test"].residual_function == "non_negative_least_squares"
    assert not m.dataset_groups["test"].link_clp


def test_global_items():
    m = Model.create_class([])(
        **{
            "clp_penalties": [
                {
                    "type": "equal_area",
                    "source": "s",
                    "source_intervals": [(1, 2)],
                    "target": "t",
                    "target_intervals": [(1, 2)],
                    "parameter": "p",
                    "weight": 1,
                }
            ],
            "clp_constraints": [
                {
                    "type": "only",
                    "target": "t",
                    "interval": [(1, 2)],
                },
                {
                    "type": "zero",
                    "target": "t",
                    "interval": (1, 2),
                },
            ],
            "clp_relations": [
                {
                    "source": "s",
                    "target": "t",
                    "interval": [(1, 2)],
                    "parameter": "p",
                },
            ],
            "dataset": {},
            "weights": [
                {"datasets": ["d1", "d2"], "value": 1},
                {"datasets": ["d3"], "value": 2, "global_interval": (5, 6)},
            ],
        }
    )
    print(m)
    assert len(m.weights) == 2
    w = m.weights[0]
    assert w.datasets == ["d1", "d2"]
    assert w.value == 1
    assert w.model_interval is None
    assert w.global_interval is None

    w = m.weights[1]
    assert w.datasets == ["d3"]
    assert w.value == 2
    assert w.model_interval is None
    assert w.global_interval == (5, 6)


def test_model_items(test_model: Model):
    assert isinstance(test_model.megacomplex["m1"], MockMegacomplexSimple)
    assert isinstance(test_model.megacomplex["m2"], MockMegacomplexWithDataset)
    assert test_model.megacomplex["m1"].dimension == "model"
    assert test_model.megacomplex["m2"].dimension == "model2"
    assert test_model.megacomplex["m1"].test_item == "t2"

    assert test_model.test_item["t1"].param == "foo"  # type:ignore[attr-defined]
    assert test_model.test_item["t1"].param_list == ["bar", "baz"]  # type:ignore[attr-defined]
    assert test_model.test_item["t1"].param_dict == {  # type:ignore[attr-defined]
        ("s1", "s2"): "baz"
    }
    assert test_model.test_item["t1"].megacomplex == "m1"  # type:ignore[attr-defined]
    assert test_model.test_item["t1"].number == 42  # type:ignore[attr-defined]
    assert test_model.test_item["t2"].param == "baz"  # type:ignore[attr-defined]
    assert test_model.test_item["t2"].param_list == ["foo"]  # type:ignore[attr-defined]
    assert test_model.test_item["t2"].param_dict == {}  # type:ignore[attr-defined]
    assert test_model.test_item["t2"].megacomplex == "m2"  # type:ignore[attr-defined]
    assert test_model.test_item["t2"].number == 7  # type:ignore[attr-defined]

    assert test_model.dataset["dataset1"].megacomplex == ["m1"]
    assert test_model.dataset["dataset1"].global_megacomplex is None
    assert test_model.dataset["dataset1"].scale == "scale_1"
    assert test_model.dataset["dataset1"].test_item_dataset == "t1"  # type:ignore[attr-defined]
    assert test_model.dataset["dataset1"].test_property_dataset1 == 1  # type:ignore[attr-defined]
    assert (
        test_model.dataset["dataset1"].test_property_dataset2 == "bar"  # type:ignore[attr-defined]
    )
    assert test_model.dataset["dataset1"].group == DEFAULT_DATASET_GROUP

    assert test_model.dataset["dataset2"].megacomplex == ["m2"]
    assert test_model.dataset["dataset2"].global_megacomplex == ["m1"]
    assert test_model.dataset["dataset2"].scale == "scale_2"
    assert test_model.dataset["dataset2"].test_item_dataset == "t2"  # type:ignore[attr-defined]
    assert test_model.dataset["dataset2"].test_property_dataset1 == 1  # type:ignore[attr-defined]
    assert (
        test_model.dataset["dataset2"].test_property_dataset2 == "bar"  # type:ignore[attr-defined]
    )
    assert test_model.dataset["dataset2"].group == "testgroup"


def test_model_as_dict():
    model_dict = {
        "clp_penalties": [
            {
                "type": "equal_area",
                "source": "s",
                "source_intervals": [(1, 2)],
                "target": "t",
                "target_intervals": [(1, 2)],
                "parameter": "p",
                "weight": 1,
            }
        ],
        "clp_constraints": [
            {"type": "only", "target": "t", "interval": [(1, 2)]},
            {"type": "zero", "target": "t", "interval": (1, 2)},
        ],
        "clp_relations": [
            {"source": "s", "target": "t", "interval": [(1, 2)], "parameter": "p"},
        ],
        "megacomplex": {
            "m1": {"type": "simple", "label": "m1", "dimension": "model", "test_item": "t1"},
        },
        "dataset_groups": {
            "default": {
                "label": "default",
                "residual_function": "non_negative_least_squares",
                "link_clp": True,
            }
        },
        "test_item": {
            "t1": {
                "label": "t1",
                "number": 4,
                "param": "foo",
                "megacomplex": "m1",
                "param_list": ["bar", "baz"],
                "param_dict": {("s1", "s2"): "baz"},
            },
        },
        "dataset": {
            "dataset1": {
                "label": "dataset1",
                "group": "default",
                "megacomplex": ["m1"],
                "global_megacomplex": ["m1"],
                "scale": "scale_1",
                "megacomplex_scale": "scale_1",
                "global_megacomplex_scale": "scale_1",
                "force_index_dependent": False,
            },
        },
        "weights": [],
    }
    as_model_dict = Model.create_class_from_megacomplexes([MockMegacomplexSimple])(
        **model_dict
    ).as_dict()
    print("want")
    print(model_dict)
    print("got")
    print(as_model_dict)
    assert as_model_dict == model_dict


def test_model_markdown(test_model: Model):
    md = test_model.markdown().replace("\n\n\n", "\n\n").replace("\n\n", "\n")
    expected = (
        (Path(__file__).parent / "expected_model_markdown.md")
        .read_text(encoding="utf8")
        .replace("\n\n\n", "\n\n")
        .replace("\n\n", "\n")
    )
    # Preprocessing to remove trailing whitespace after '* *Matrix*:'
    expected = "\n".join([line.rstrip(" ") for line in str(expected).split("\n")])
    result = "\n".join([line.rstrip(" ") for line in str(md).split("\n")])
    assert result == expected


def test_get_parameter_labels(test_model: Model):
    wanted = {"foo", "scale_1", "scale_2", "baz", "bar"}
    got = test_model.get_parameter_labels()
    print(got)
    assert wanted == got



from glotaran.cli.commands.util import project_io_list_supporting_plugins


def test_project_io_list_supporting_plugins_save_result():
    """Same as used in ``--outformat`` CLI option."""
    result = project_io_list_supporting_plugins("save_result", ("yml_str"))

    assert "csv" not in result
    assert "yml_str" not in result
    assert "folder" in result



"""The glotaran parameter history package."""
from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

import numpy as np
import pandas as pd

from glotaran.parameter.parameters import Parameters

if TYPE_CHECKING:
    from os import PathLike


class ParameterHistory:
    """A class representing a history of parameters."""

    def __init__(self):  # noqa: D107
        self._parameter_labels: list[str] = []
        self._parameters: list[np.ndarray] = []
        self.source_path = "parameter_history.csv"

    @classmethod
    def from_dataframe(cls, history_df: pd.DataFrame) -> ParameterHistory:
        """Create a history from a pandas data frame.

        Parameters
        ----------
        history_df : pd.DataFrame
            The source data frame.

        Returns
        -------
        ParameterHistory
            The created history.
        """
        history = cls()

        history._parameter_labels = history_df.columns

        for parameter_values in history_df.values:
            history._parameters.append(parameter_values)

        return history

    @classmethod
    def from_csv(cls, path: str) -> ParameterHistory:
        """Create a history from a csv file.

        Parameters
        ----------
        path : str
            The path to the csv file.

        Returns
        -------
        ParameterHistory
            The created history.
        """
        df = pd.read_csv(path)
        return cls.from_dataframe(df)

    loader = from_csv

    @property
    def parameter_labels(self) -> list[str]:
        """Return the labels of the parameters in the history.

        Returns
        -------
        list[str]
            A list of parameter labels.
        """
        return self._parameter_labels

    @property
    def parameters(self) -> list[np.ndarray]:
        """Return the parameters in the history.

        Returns
        -------
        list[np.ndarray]
            A list of parameters in the history.
        """
        return self._parameters

    def __len__(self) -> int:
        """Return the number of records in the history."""
        return self.number_of_records

    @property
    def number_of_records(self) -> int:
        """Return the number of records in the history.

        Returns
        -------
        int
            The number of records.
        """
        return len(self._parameters)

    def to_dataframe(self) -> pd.DataFrame:
        """Create a data frame from the history.

        Returns
        -------
        pd.DataFrame
            The created data frame.
        """
        return pd.DataFrame(self._parameters, columns=self.parameter_labels)

    def to_csv(self, file_name: str | PathLike[str], delimiter: str = ","):
        """Write a :class:`ParameterHistory` to a CSV file.

        Parameters
        ----------
        file_name : str
            The path to the CSV file.
        delimiter : str
            The delimiter of the CSV file.
        """
        self.source_path = Path(file_name).as_posix()
        self.to_dataframe().to_csv(file_name, sep=delimiter, index=False)

    def append(self, parameters: Parameters, current_iteration: int = 0):
        """Append :class:`Parameters` to the history.

        Parameters
        ----------
        parameters : Parameters
            The group to append.
        current_iteration: int
            Current iteration of the optimizer.

        Raises
        ------
        ValueError
            Raised if the parameter labels differs from previous.
        """
        (
            parameter_labels,
            parameter_values,
            _,
            _,
        ) = parameters.get_label_value_and_bounds_arrays()
        parameter_labels = ["iteration", *parameter_labels]
        if len(self._parameter_labels) == 0:
            self._parameter_labels = parameter_labels
        if parameter_labels != self.parameter_labels:
            raise ValueError("Cannot append parameters. Parameter labels do not match existing.")

        self._parameters.append(np.array([current_iteration, *parameter_values]))

    def get_parameters(self, index: int) -> np.ndarray:
        """Get parameters for a history index.

        Parameters
        ----------
        index : int
            The history index.

        Returns
        -------
        np.ndarray
            The parameter values at the history index as array.
        """
        return self._parameters[index]



import numpy as np

from glotaran.parameter.parameter_history import ParameterHistory
from glotaran.parameter.parameters import Parameters


def test_parameter_history():
    group0 = Parameters.from_list([["1", 1], ["2", 4]])
    group1 = Parameters.from_list([["1", 2], ["2", 5]])
    group2 = Parameters.from_list([["1", 3], ["2", 6]])

    history = ParameterHistory()

    history.append(group0)

    assert history.parameter_labels == ["iteration", "1", "2"]

    assert history.number_of_records == 1
    assert all(history.get_parameters(0) == [0, 1, 4])

    history.append(group1, current_iteration=1)

    assert history.number_of_records == 2
    assert all(history.get_parameters(1) == [1, 2, 5])

    history.append(group2, current_iteration=2)

    assert history.number_of_records == 3
    assert all(history.get_parameters(2) == [2, 3, 6])

    df = history.to_dataframe()

    assert all(df.columns == history.parameter_labels)
    assert np.all(df.values == history.parameters)

    group2.set_from_history(history, 0)

    assert group2.get("1") == 1
    assert group2.get("2") == 4



"""This module contains clp penalty items."""

from __future__ import annotations

from glotaran.model.item import ParameterType
from glotaran.model.item import TypedItem
from glotaran.model.item import item


@item
class ClpPenalty(TypedItem):
    """Baseclass for clp penalties."""


@item
class EqualAreaPenalty(ClpPenalty):
    """Forces the area of 2 clp to be the same.

    An equal area constraint adds a the difference of the sum of a
    compartments in the e matrix in one or more intervals to the scaled sum
    of the e matrix of one or more target compartments to residual. The additional
    residual is scaled with the weight.
    """

    type: str = "equal_area"
    source: str
    source_intervals: list[tuple[float, float]]
    target: str
    target_intervals: list[tuple[float, float]]
    parameter: ParameterType
    weight: float



"""Glotaran package root."""
from glotaran.deprecation.deprecation_utils import deprecate_submodule
from glotaran.plugin_system.base_registry import load_plugins

load_plugins()

__version__ = "0.7.0"

examples = deprecate_submodule(
    deprecated_module_name="glotaran.examples",
    new_module_name="glotaran.testing.simulated_data",
    to_be_removed_in_version="0.8.0",
    module_load_overwrite="glotaran.deprecation.modules.examples",
)



from copy import deepcopy

import pytest

from glotaran.model import ZeroConstraint
from glotaran.optimization.optimization_group import OptimizationGroup
from glotaran.optimization.test.suites import TwoCompartmentDecay as suite
from glotaran.project import Scheme
from glotaran.simulation import simulate


@pytest.mark.parametrize("index_dependent", [True, False])
@pytest.mark.parametrize("link_clp", [True, False])
def test_constraint(index_dependent, link_clp):
    model = deepcopy(suite.model)
    model.dataset_groups["default"].link_clp = link_clp
    model.megacomplex["m1"].is_index_dependent = index_dependent
    model.clp_constraints.append(ZeroConstraint(**{"target": "s2", "interval": (1, 1)}))

    print("link_clp", link_clp, "index_dependent", index_dependent)
    dataset = simulate(
        suite.sim_model,
        "dataset1",
        suite.wanted_parameters,
        {"global": suite.global_axis, "model": suite.model_axis},
    )
    scheme = Scheme(model=model, parameters=suite.initial_parameters, data={"dataset1": dataset})
    optimization_group = OptimizationGroup(scheme, model.get_dataset_groups()["default"])
    optimization_group.calculate(suite.initial_parameters)

    reduced_matrix = (
        optimization_group._matrix_provider.get_aligned_matrix_container(0)
        if link_clp
        else optimization_group._matrix_provider.get_prepared_matrix_container("dataset1", 0)
    )
    matrix = optimization_group._matrix_provider.get_matrix_container("dataset1")

    result_data = optimization_group.create_result_data()
    print(result_data)  # T201
    clps = result_data["dataset1"].clp

    assert "s2" not in reduced_matrix.clp_labels
    assert "s2" in clps.coords["clp_label"]
    assert clps.sel(clp_label="s2")[0] == 0
    assert clps.sel(clp_label="s2")[1] != 0
    assert "s2" in matrix.clp_labels

    # 1 compartment * 2 items in global axis
    # + 1 compartment * 1 item in global axis
    assert optimization_group.number_of_clps == 3
